import pandas as pd
import os
import numpy as np
import math

from sklearn.decomposition import NMF

from adlinear import pca
from adlinear import utilities as utl
from adlinear import nmfmodel as nmf
from adlinear import testers as tst
from adlinear import imputer as imp
from adlinear import clusterizer as clu
from adlinear import ntfmodel as ntf
from randomgenerators import randomgenerators as rng

import root_path
import dotenv
from pathlib import Path
dotenv.load_dotenv()


if __name__ == "__main__":

    root_path = root_path.get_root_path()

    rd_path = root_path / os.getenv("rd_subpath")
    imputation_paper_path = rd_path / os.getenv("imputation_paper_subpath")

    data_path = imputation_paper_path / os.getenv("data_subpath")

    temp_data_path = data_path / "temp_data/"
    pictures_path = data_path / "Pics"
    out_data_path = imputation_paper_path / "results/"

    res_path = imputation_paper_path / os.getenv("results_subpath")

    if os.getenv("imp_do_brunet", "False").lower() == "true":
        df_brunet = pd.read_csv(data_path / os.getenv("brunet_filename"))
        df_brunet_groups = pd.read_csv(data_path / os.getenv("brunet_grp_filename"))
        df_brunet.loc[:, "Group"] = df_brunet_groups.loc[:, "Group"]
        ngrps_brunet = 3
        ncomp_brunet = 4
        # la colonne d'attribution des groupes
        if ngrps_brunet == 4:
            df_brunet_priors = df_brunet.replace({"Group": {"ALL_B_1": 1,
                                                            "ALL_B_2": 2,
                                                            "ALL_T": 3,
                                                            "AML": 4}})["Group"]
        else:
            df_brunet_priors = df_brunet.replace({"Group": {"ALL_B": 1,
                                                            "ALL_T": 2,
                                                            "AML": 3}})["Group"]

        df_brunet = df_brunet[df_brunet.columns[2:]]
        df_log_brunet = df_brunet - np.nanmin(df_brunet, axis=0)
        df_log_brunet = np.log(1.0 + df_log_brunet)

        nmf_brunet = nmf.NmfModel(mat=df_brunet, name="brunet",
                                  ncomp=ncomp_brunet, regularization="components", leverage="robust", max_iter=200)
        nmf_log_brunet = nmf.NmfModel(mat=df_log_brunet, name="log_brunet",
                                      ncomp=ncomp_brunet, regularization="components", leverage="robust", max_iter=200)

        if os.getenv("imp_do_brunet_test_prop", "False").lower() == "true":
            # df_res_brunet = tst.test_missing_proportion(df_log_brunet, ncomp=ncomp_brunet,
            #                                             name="log_brunet", grp_priors=df_brunet_priors,
            #                                             p_min=0.0, p_max=0.4, p_step=0.02, n_trials=50,
            #                                             do_save_result=True, outpath=out_data_path)

            p_step = 0.05
            n_step = int(0.4 / p_step)
            missing_props = [round(i * p_step, 2) for i in range(n_step)]
            # missing_props = [0]

            clusters_brunet_priors = clu.Clusterizer(method="set_groups",
                                                     groups=df_brunet_priors,
                                                     name="Brunet_3_groups",
                                                     nb_groups=ngrps_brunet)
            clusters_nmf4_grp3 = clu.Clusterizer(method="nmf.kmeans", nb_groups=3, ncomp=4)

            imp_kmeans = imp.Imputer(method="kmeans",
                                     params={"ngroups": 3})
            imp_nmf_proxy = imp.Imputer(method="nmf.proxy",
                                        params={"ncomp": ncomp_brunet,
                                                "nfill_iters": 0})
            imp_snmf_proxy = imp.Imputer(method="snmf.proxy",
                                         params={"ncomp": ncomp_brunet,
                                                 "nblocks": 2,
                                                 "nfill_iters": 0})

            imp_mean = imp.Imputer("mean", params={})

            imputers = [imp_mean, imp_kmeans, imp_nmf_proxy, imp_snmf_proxy]
            # imputers = [imp_mean, imp_kmeans]

            # imp_tester = imp.ImputerTester(mat=df_log_brunet, name="Log_Brunet_MisClass", imputer=imp_mean,
            #                                ref_clst=clusters_brunet_priors, clst=clusters_nmf4_grp3,
            #                                err_func="Misclassifieds")
            # # missing_props = [0, 0.05]
            # for imputer in imputers:
            #     imp_tester.set_imputer(imputer)
            #     imp_tester.run(missing_props, 100)

            # imp_tester.output_results(out_data_path)

            imp_tester = imp.ImputerTester(mat=df_log_brunet, name="Log_Brunet_MSE", imputer=imp_mean,
                                           ref_clst=clusters_brunet_priors, clst=clusters_nmf4_grp3, err_func="MSE")
            for imputer in imputers:
                imp_tester.set_imputer(imputer)
                imp_tester.run(missing_props, 100)

            imp_tester.output_results(out_data_path)

    if os.getenv("imp_do_random", "False").lower() == "true":

        standard_noise = rng.RandomVariable(np.random.normal, 1, 0.05)

        # direct_x_clones, _ = standard_noise.apply_bias(size=[10, 100], min=0.0, max=4.0)
        # biased_x_clones, _ = standard_noise.apply_bias(size=[10, 100], signal_prop=0.1, bias=0.1, min=0.0, max=4.0)

        dict1 = {"Variable": standard_noise,
                 "Signal_prop": 0.10,
                 "Bias": 0.2,
                 "Coeff": 0.01,
                 "Min": -10,
                 "Max": 10
                 }
        dict2 = {"Variable": standard_noise,
                 "Signal_prop": 0.05,
                 "Bias": 0.1,
                 "Coeff": 0.01,
                 "Min": -10,
                 "Max": 10
                 }
        dict3 = {"Variable": standard_noise,
                 "Signal_prop": 0.20,
                 "Bias": -0.2,
                 "Coeff": 0.01,
                 "Min": -10,
                 "Max": 10
                 }
        dict4 = {"Variable": standard_noise,
                 "Signal_prop": 0.40,
                 "Bias": -0.4,
                 "Min": -10,
                 "Max": 10
                 }
        noise = {"Variable": standard_noise,
                 "Signal_prop": 0.0,
                 "Bias": 0.0,
                 "Coeff": 0.01,
                 "Min": 0,
                 "Max": 4
                 }
        var_list = [dict1, dict2, dict3, dict4]
        noise_factor = 0.50
        n_clones = 100
        n_samples = 250
        my_dls = rng.DependentLocalizedSignals(signal_dist=var_list,
                                               non_overlapping_obs=True,
                                               noise_dist=noise,
                                               cloning_mult=n_clones,
                                               n_crossproducts=30,
                                               n_noisecolumns=int(n_clones*noise_factor),
                                               nsamples=250,
                                               lbound=-10,
                                               ubound=10)
        my_dls_name = my_dls.__repr__()
        _ = my_dls()
        df_rnd_samples = my_dls.get_samples()

        rnd_ngroups = len(var_list) + 1
        df_rnd_groups = df_rnd_samples.loc[:, "Group"]
        df_rnd_mat = df_rnd_samples.drop("Group", axis=1)

        p_step = 0.05
        n_step = int(0.4 / p_step)
        missing_props = [round(i * p_step, 2) for i in range(n_step)]
        # missing_props = [0]

        clusters_rnd_priors = clu.Clusterizer(method="set_groups",
                                              groups=df_rnd_groups,
                                              name=f"{my_dls_name}_{rnd_ngroups}_groups",
                                              nb_groups=rnd_ngroups)

        clusters_nmf = clu.Clusterizer(method="nmf.kmeans", nb_groups=rnd_ngroups, ncomp=rnd_ngroups)
        clusters_kmeans = clu.Clusterizer(method="kmeans", nb_groups=rnd_ngroups, ncomp=rnd_ngroups)

        imp_kmeans = imp.Imputer(method="kmeans",
                                 params={"ngroups": rnd_ngroups})
        imp_nmf_proxy = imp.Imputer(method="nmf.proxy",
                                    params={"ncomp": rnd_ngroups,
                                            "nfill_iters": 0})
        imp_nmf_it1_proxy = imp.Imputer(method="nmf.proxy",
                                        params={"ncomp": rnd_ngroups,
                                                "nfill_iters": 1})

        imp_snmf_proxy = imp.Imputer(method="snmf.proxy",
                                     params={"ncomp": rnd_ngroups,
                                             "nblocks": 2,
                                             "nfill_iters": 0})
        imp_snmf_it1_proxy = imp.Imputer(method="snmf.proxy",
                                         params={"ncomp": rnd_ngroups,
                                                 "nblocks": 2,
                                                 "nfill_iters": 1})

        imp_nmf_fills = imp.Imputer(method="nmf.proxy", params={"ncomp": rnd_ngroups, "nfill_iters": 4})
        imp_snmf_proxy_fills = imp.Imputer(method="snmf.proxy",
                                           params={"ncomp": rnd_ngroups, "nblocks": 2, "nfill_iters": 4})
        imp_mean = imp.Imputer("mean", params={})

        imputers = [[imp_mean, clusters_kmeans],
                    [imp_kmeans, clusters_kmeans],
                    [imp_snmf_proxy, clusters_kmeans],
                    [imp_snmf_proxy, clusters_nmf]
                    ]

        # imputers = [[imp_mean, clusters_kmeans]]
        # imputers = [[imp_kmeans, clusters_kmeans]]

        imp_tester = imp.ImputerTester(mat=df_rnd_mat, name=my_dls_name, imputer=imp_mean, ref_clst=clusters_rnd_priors,
                                       clst=clusters_nmf, err_func="MSE")
        for imputer in imputers:
            imp_tester.set_imputer(imputer[0])
            imp_tester.set_clst(imputer[1])
            imp_tester.run(missing_props, 100)

        imp_tester.output_results(out_data_path)

    if os.getenv("imp_do_random_block_matrices", "False").lower() == "true":
        # Expérience pour prouver que l'imputation de valeurs via SNMF
        # donne les meilleurs résultats quand on connait le nombre exact de composants
        # nb_clusters = np.random.randint(low=5, high=40)
        nb_clusters = 12
        min_corr = np.random.uniform(low=0.75, high=0.95)
        min_corr = 0.90
        max_corr = np.random.uniform(low=0.05, high=0.25)
        max_corr = 0.10
        h_avg_csize = np.random.randint(low=10, high=20)
        h_avg_csize = 20
        w_avg_csize = np.random.randint(low=10, high=100)
        w_avg_csize = 40
        rnd_norms = False
        # eps = np.random.uniform(low=0.0, high=0.25)
        eps = 0.0
        generated_M = rng.generate_nmf_reconstruction(n_comp=nb_clusters,
                                                      min_intra_corr=min_corr,
                                                      max_inter_corr=max_corr,
                                                      h_avg_clust_size=h_avg_csize,
                                                      w_avg_clust_size=w_avg_csize,
                                                      epsilon=eps,
                                                      random_norms=rnd_norms)
        imp_nmf_proxy = imp.Imputer(method="nmf.proxy",
                                    params={"ncomp": 3,
                                            "nblocks": 1,
                                            "nfill_iters": 0})

        imp_nmf_proxy_pgrad = imp.Imputer(method="nmf.proxy",
                                          params={"ncomp": 3,
                                                  "nblocks": 1,
                                                  "use_hals": False,
                                                  "nfill_iters": 0})

        imputers = [[imp_nmf_proxy, None],
                    [imp_nmf_proxy_pgrad, None]]
        imp_name = "RandomBlockMatrices_2"
        imp_tester = imp.ImputerTester(mat=generated_M, name=f"{imp_name}_nc{nb_clusters}",
                                       imputer=imp_nmf_proxy, ref_clst=None,
                                       clst=None, err_func="MSE")
        for imp in imputers:
            for nc in range(5, 20):
                imputer = imp[0]
                imputer.set_ncomp(nc)
                imp_tester.set_imputer(imputer)
                # imp_tester.run([0.05, 0.10, 0.25, 0.4], ntrials=20)
                imp_tester.run([0.05], ntrials=1)

        imp_tester.output_results(out_data_path)

    if os.getenv("imp_do_screeplot_tests", "False").lower() == "true":
        # wh_name = os.getenv("imp_wh_matname")
        # wh_mat = pd.read_csv(data_path / wh_name)
        # wh_mat.drop(wh_mat.columns[0], axis='columns')
        # wh_screeplot = nmf.generate_scree_plot(wh_mat, 2, 20)

        nb_clusters = 10
        min_corr = 0.9
        max_corr = 0.10
        h_avg_csize = 20
        w_avg_csize = 20
        eps = 0
        random_norms = True
        censor_rate = 0.01
        read_wh_mat = True
        rnstr = "RandNorms" if random_norms else "ConstNorms"

        for censor_rate in [0.0, 0.01, 0.05, 0.10]:

            if read_wh_mat:
                fixed_mat_name = os.getenv("imp_wh_matname")
                rad_name = fixed_mat_name.split(".csv")[0]
                wh_name = f"{rad_name}_mis{censor_rate}.csv"
                # wh_name = "M_nc10_corrmin0.9_corrmax0.1_noise0_RandNorms_mis0.0.csv"
                generated_M = pd.read_csv(data_path / fixed_mat_name, index_col=0)
            else:
                wh_name = f"M_nc{nb_clusters}_corrmin{round(min_corr, 2)}_" \
                          f"corrmax{round(max_corr, 2)}_noise{round(eps, 2)}_{rnstr}_mis{censor_rate}.csv"
                generated_M = rng.generate_nmf_reconstruction(n_comp=nb_clusters,
                                                              min_intra_corr=min_corr,
                                                              max_inter_corr=max_corr,
                                                              h_avg_clust_size=h_avg_csize,
                                                              w_avg_clust_size=w_avg_csize,
                                                              epsilon=eps,
                                                              random_norms=random_norms)

            utl.censor_data(mat=generated_M, p=censor_rate, inplace=True)

            generated_M.to_csv(res_path / "random_nmf" / wh_name,
                               float_format="%.4f")
            df_scree_plot = nmf.generate_scree_plot(generated_M, ncmin=6, ncmax=30)
            df_scree_plot.to_csv(res_path / f"{wh_name}_screeplot.csv")

    if os.getenv("imp_do_ecodata_gans", "False").lower() == "true":
        # expérience où l'on soumet les résultats de l'imputation à un GAN
        # localisation du modèle GAN
        path_models = Path('/media/SERVEUR/production/research_and_development/AdFactory/Models/')
        # model_name = '15_09_21_25000_epochs'
        model_name = '20_09_21_10000_epochs'
        path_output_models = path_models / model_name
        critic_name = 'wgan_critic_model'

        eco_feat_list = ['Balance_of_Trade', 'Central_Bank_Balance_Sheet', 'Corruption_Index', 'Food_Inflation',
                         'Foreign_Direct_Investment', 'GDP', 'GDP_Growth_Rate',
                         'Population', 'Terrorism_Index', 'BONDS_10Y_close']
        nb_eco_feat = len(eco_feat_list)
        nb_obs = 500

        # Données d'entrée
        eco_data_path = Path('/media/SERVEUR/production/research_and_development/AdFactory/Data/test_critic_full')
        eco_df = pd.DataFrame(index=range(nb_obs),
                              columns=eco_feat_list,
                              data=np.random.normal(size=[nb_obs, nb_eco_feat]))

        ncomp_eco = 6
        p_step = 0.05
        n_step = int(0.4 / p_step)
        missing_props = [round(i * p_step, 2) for i in range(n_step)]
        imp_kmeans = imp.Imputer(method="kmeans",
                                 params={"ngroups": ncomp_eco})
        imp_nmf_proxy = imp.Imputer(method="nmf.proxy",
                                    params={"ncomp": ncomp_eco,
                                            "nfill_iters": 0})
        imp_snmf_proxy = imp.Imputer(method="snmf.proxy",
                                     params={"ncomp": ncomp_eco,
                                             "nblocks": 2,
                                             "nfill_iters": 0})
        clusters_kmeans = clu.Clusterizer(method="kmeans", nb_groups=ncomp_eco, ncomp=ncomp_eco)

        imputers = [[imp_kmeans, clusters_kmeans],
                    [imp_snmf_proxy, clusters_kmeans],
                    ]

        imp_tester = imp.ImputerTester(mat=eco_df, name="Trading_eco", imputer=imp_kmeans, ref_clst=None, clst=None,
                                       err_func="GAN_critic", critic_path=path_output_models, critic_name=critic_name)

        p_step = 0.05
        n_step = int(0.4 / p_step)
        missing_props = [round(i * p_step, 2) for i in range(n_step)]

        file_list = [f"df_{str(1000+i)[1:4]}.csv" for i in range(1, 101)]
        i_choice = range(5)
        for i in i_choice:
            mat = pd.read_csv(eco_data_path / file_list[i])
            mat = mat.drop(["Date", "Countries"], axis=1)
            imp_tester.set_mat(mat)
            imp_tester.set_name(file_list[i])
            for imputer in imputers:
                imp_tester.set_imputer(imputer[0])
                clusterizer = imputer[1]
                score_opt, scores = clu.clusterizer_optimize(clusterizer, mat)
                imp_tester.set_clst(clusterizer)

                imp_tester.run(missing_props, 100)

        imp_tester.output_results(out_data_path)

        pass



import numpy as np
from adlinear import utilities as utl

epsilon = 1e-10


def almost_eq(x, y):
    return abs(x-y) <= epsilon


def test_hhi():
    res = utl.hhi_1([0, 1, 0, 1, 0, 1])
    assert (almost_eq(res, 3))
    res = utl.hhi_1([-1] * 100)
    assert (almost_eq(res, 100))
    res = utl.hhi_1([0])
    assert (res == 0)


def test_mean_square():

    for i in range(2, 10):
        mytab = np.arange(start=1, stop=i+1)
        res = utl.mean_square(mytab)
        assert (almost_eq(res, (i+1)*(2*i+1)/6.0))



import math
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import scipy.stats

# import typing
from typing import Union, Tuple
from sklearn.cluster import KMeans
from sklearn import metrics
# from . import clusterizer as cls


def hhi_1(tab) -> float:
    tab = [abs(x) for x in tab]
    listsum = sum(tab)
    if listsum == 0:
        return 0
    else:
        rtab = [x / listsum for x in tab]
        tab2 = [x * x for x in rtab]
        return 1.0 / sum(tab2)


def mat_hhi_1(mat, axis=0) -> pd.DataFrame:
    mat = pd.DataFrame(mat)
    if axis == 0:
        res = pd.DataFrame(index=mat.columns, columns=["InvHHI"])
        res.loc[:, "InvHHI"] = [hhi_1(mat.loc[:, col]) for col in mat.columns]
    else:
        res = pd.DataFrame(index=mat.index, columns=["InvHHI"])
        res.loc[:, "InvHHI"] = [hhi_1(mat.loc[row, :]) for row in mat.index]
    return res


def count_not_nans(mat) -> int:
    return np.count_nonzero(~np.isnan(mat.astype(float)))


def count_nans(mat) -> int:
    return np.count_nonzero(np.isnan(mat.astype(float)))


def mean_square(mat) -> float:
    sz = count_not_nans(mat)
    if sz == 0:
        return np.nan
    else:
        mat = np.nan_to_num(mat, nan=0.0)
        return np.sum(mat ** 2) / sz


def norm_n2(vec) -> float:
    vec = np.nan_to_num(vec, nan=0.0)
    return (np.sum(vec**2))**0.5


def relative_squared_error(mat1, mat2, refmat=None) -> float:
    assert mat1.shape == mat2.shape
    if refmat is None:
        meansq1 = mean_square(mat1)
        meansq2 = mean_square(mat2)
        meansq = 0.5 * (meansq1 + meansq2)
    else:
        meansq = mean_square(refmat)
    return mean_square(mat1 - mat2) / meansq if meansq > 0 else 1e9


def freq_bins(mat, nbins=10) -> Union[tuple, np.ndarray, float]:
    """

     freq_bins:

     Parameters
     ----------

     mat: a float array or matrix

     nbins: a positive integer representing a linear binning of the values

     :returns:
     a DataFrame containing the features (attributes) as columns, the observations (e.g stocks) as lines
     """

    values = pd.DataFrame(np.ravel(mat)).dropna()
    return scipy.stats.relfreq(values, nbins)


def normalized_entropy(mat, nbins=3) -> float:
    assert nbins > 0
    if nbins == 1:
        return 0
    else:
        freqs = freq_bins(mat, nbins)
        return scipy.stats.entropy(freqs.frequency) / math.log(nbins)


def relative_entropy_by_cluster(mat, lineclusters, colclusters, nbins=10) :
    assert nbins > 0
    nl, nc = mat.shape
    assert nl == len(lineclusters)
    assert nc == len(colclusters)
    l_clust_values = pd.unique(pd.DataFrame(lineclusters).iloc[:, 0].sort_values())
    c_clust_values = pd.unique(pd.DataFrame(colclusters).iloc[:, 0].sort_values())

    clust_values = list(set(l_clust_values).intersection(c_clust_values))
    df_entropies = pd.DataFrame(index=clust_values, columns=clust_values)
    df_sizes = pd.DataFrame(index=clust_values, columns=clust_values)
    for l_val in clust_values:
        for c_val in clust_values:
            val_lines = lineclusters[lineclusters == l_val].dropna().index
            val_cols = colclusters[colclusters == c_val].dropna().index
            submat = mat.loc[val_lines, val_cols]
            df_entropies.loc[l_val, c_val] = normalized_entropy(submat, nbins)
            df_sizes.loc[l_val, c_val] = len(val_lines) * len(val_cols)
    return df_entropies, df_sizes


def clustering_caha_score(mat, clusters):
    return metrics.calinski_harabasz_score(mat, clusters)


def clustering_caha_score_by_ncomp(mat, clustermodel, ncompmin=2, ncompmax=20):
    kmeans_model = clustermodel.fit(mat)
    return metrics.calinski_harabasz_score(mat, clustermodel)


def censor_data(mat, p=0.0, inplace=True, randomgen=None):
    """

    """
    assert (0.0 <= p < 1.0)
    if p == 0.0:
        return mat.copy() if not inplace else None
    else:
        if randomgen is None:
            randomgen = np.random.default_rng()
        nr = len(mat.index)
        nans = int(abs(p) * nr)
        matc = mat if inplace else mat.copy()

        for icol, _ in enumerate(matc.columns):
            # censored_idx = np.random.Generator.choice(range(nr), nans, replace=False)
            censored_idx = randomgen.choice(range(nr), nans, replace=False)
            matc.iloc[censored_idx, icol] = np.nan

        return None if inplace else matc


def swap_in_mat(mat, i, j, axis=0, inplace=True):
    """
    swap_in_mat: swaps two lines or two rows of a matrix.
    Can be in place or not.

    :returns
    None if in place, swapped matrix otherwise
    """
    nr, nc = mat.shape
    assert (0 <= i < nr and 0 <= j < nc)
    if i == j:
        return None if inplace else mat
    else:
        matc = mat if inplace else mat.copy()
        chunk_i = mat.iloc[i, :] if axis == 0 else mat.iloc[:, i]
        chunk_j = mat.iloc[j, :] if axis == 0 else mat.iloc[:, j]
        temp = chunk_i.copy()
        if axis == 0:
            matc.iloc[i, :] = chunk_j
            matc.iloc[j, :] = temp
        else:
            matc.iloc[:, i] = chunk_j
            matc.iloc[:, j] = temp

        return None if inplace else matc


def make_diagonal_dominant(mat, inplace=True):
    nr, nc = mat.shape
    assert (nr == nc)
    matc = mat if inplace else mat.copy()
    for ir, _ in enumerate(mat.index):
        cmax = np.argmax(matc.iloc[ir, :])
        swap_in_mat(matc, ir, cmax, axis=1, inplace=True)
    return None if inplace else matc


def get_observations_km_clusters(mat, ngroups=2):

    km = KMeans(n_clusters=ngroups)
    return km.fit_predict(mat)


def get_clustering_incidence_matrix(mat: Union[np.ndarray, pd.DataFrame],
                                    df_clust1: Union[np.ndarray, pd.DataFrame, Tuple[int]],
                                    df_clust2: Union[np.ndarray, pd.DataFrame, Tuple[int]]):

    assert (len(mat) == len(df_clust1) and len(df_clust1) == len(df_clust2))
    df_clust1 -= np.min(df_clust1) - 1
    df_clust2 -= np.min(df_clust2) - 1
    nclust1 = np.max(df_clust1)
    nclust2 = np.max(df_clust2)
    incidence_mat = pd.DataFrame(index=sorted(pd.unique(df_clust1.iloc[:, 0])),
                                 columns=sorted(pd.unique(df_clust2.iloc[:, 0]))).fillna(0.0, inplace=False)
    df_clust1 = pd.DataFrame(df_clust1)
    df_clust2 = pd.DataFrame(df_clust2)
    for iobs, obs in enumerate(mat.index):
        clust1 = int(df_clust1.iloc[iobs])
        clust2 = int(df_clust2.iloc[iobs])
        incidence_mat.loc[clust1, clust2] = incidence_mat.loc[clust1, clust2] + 1

    return incidence_mat


def get_clustering_mismatches(mat: Union[np.ndarray, pd.DataFrame],
                              df_clust1: Union[np.ndarray, pd.DataFrame, Tuple[int]],
                              df_clust2: Union[np.ndarray, pd.DataFrame, Tuple[int]]):
    inc_mat = get_clustering_incidence_matrix(mat, df_clust1, df_clust2)
    make_diagonal_dominant(inc_mat, inplace=True)
    nobj = len(mat.index)
    nmis = nobj - np.trace(inc_mat)
    return nmis, float(nmis) / nobj


def get_nb_misclassifieds_old(mat, grp_priors):
    ngrps = len(set(grp_priors))
    km_grps = get_observations_km_clusters(mat, ngrps)
    return get_clustering_mismatches(mat, km_grps, grp_priors)


def get_nb_misclassifieds(mat: Union[np.ndarray, pd.DataFrame],
                          cltzr,
                          ref_cltzr):
    ngrps = cltzr.get_nbgroups()
    ref_ngrps = ref_cltzr.get_nbgroups()
    assert(ngrps == ref_ngrps)
    grps = cltzr.apply(mat)
    ref_grps = ref_cltzr.apply(mat)
    return get_clustering_mismatches(mat, grps, ref_grps)


def get_avg_features_by_clusters(mat: Union[np.ndarray, pd.DataFrame], cltzr):
    groups = cltzr.apply(mat)
    ngroups = len(set(groups))
    df_mat = pd.DataFrame(mat)
    df_means = pd.DataFrame(index=range(ngroups), columns=df_mat.columns)
    df_mat['cluster'] = groups
    for grp in range(ngroups):
        submat = df_mat[df_mat['cluster'] == grp]
        df_means.loc[grp, :] = np.nanmean(submat.drop(['cluster'], axis=1), axis=0)
    return df_means, groups


def get_avg_features_by_km_clusters(mat: Union[np.ndarray, pd.DataFrame], ngroups=2):
    km = get_observations_km_clusters(mat, ngroups)
    df_mat = pd.DataFrame(mat)
    df_means = pd.DataFrame(index=range(ngroups), columns=df_mat.columns)
    df_mat['cluster'] = km
    for grp in range(ngroups):
        submat = df_mat[df_mat['cluster'] == grp]
        df_means.loc[grp, :] = np.nanmean(submat.drop(['cluster'], axis=1), axis=0)
    return df_means, km


def fill_missing(mat, method="median", axis=0, fill_values=None, ngroups=0):

    if type(mat) is not pd.DataFrame:
        cols = list(["F" + str(x) for x in range(1, mat.shape[1] + 1)])
        filled_mat = pd.DataFrame(data=mat.values, columns=cols)
    else:
        filled_mat = mat.copy().astype(float)

    if method in ["median", "mean"]:
        if method == "median":
            df_estimates = mat.astype(float).apply(np.nanmedian, axis=axis)
        elif method == "mean":
            df_estimates = mat.astype(float).apply(np.nanmean, axis=axis)
        if axis == 0:
            # TODO: do the fill in next step without for loop
            # for col in mat.columns:
            #     filled_mat[col].fillna(value=df_estimates[col], inplace=True)
            filled_mat.fillna(value=df_estimates.to_dict(), inplace=True)
        else:
            for idx in mat.index:
                filled_mat.loc[idx].fillna(value=df_estimates.loc[idx], inplace=True)

    elif method == "kmeans":
        # KMeans ne tolère pas les valeurs manquantes: imputer la moyenne par feature
        km_filled_mat = fill_missing(mat, method="mean", axis=axis)
        df_means, km = get_avg_features_by_km_clusters(km_filled_mat, ngroups)
        # reprendre la matrice incomplète
        filled_mat = mat.copy()
        for idx in filled_mat.index:
            grp = km[idx]
            filled_mat.loc[idx] = mat.loc[idx].fillna(value=df_means.loc[grp, :])

    elif method == "values" and fill_values is not None:
        # update works inplace
        filled_mat.update(other=fill_values, overwrite=False)

    return filled_mat


def shorten_features_names(mat, wished_len=20):
    if hasattr(mat, "columns"):
        cols_dict = dict.fromkeys(list(mat.columns))
        for col in cols_dict:
            cols_dict[col] = col[0:wished_len] if len(col) >= wished_len else col
        mat = mat.rename(columns=cols_dict)
    return mat


def center_by_columns(mat):
    if type(mat) is pd.DataFrame:
        pass
    else:
        cols = list(["F" + str(x) for x in range(1, mat.shape[1] + 1)])
        mat = pd.DataFrame(data=mat.values, columns=cols)
    centered_mat = mat.copy()
    return centered_mat


def normalize_by_columns(mat):
    if type(mat) is not pd.DataFrame:
        cols = list(["F" + str(x) for x in range(1, mat.shape[1] + 1)])
        mat = pd.DataFrame(data=mat.values, columns=cols)
    normalized_mat = mat.copy()
    means = np.nanmean(normalized_mat, axis=0)
    stdevs = np.nanstd(normalized_mat, axis=0)
    normalized_mat -= means
    normalized_mat /= stdevs

    return normalized_mat, means, stdevs


def denormalize_by_columns(mat, means=0, stdevs=1):
    if type(mat) is not pd.DataFrame:
        cols = list(["F" + str(x) for x in range(1, mat.shape[1] + 1)])
        mat = pd.DataFrame(data=mat.values, columns=cols)
    denormalized_mat = mat.copy()
    denormalized_mat *= stdevs
    denormalized_mat += means

    return denormalized_mat


def sort_columns_by_fill_rate(mat, grouping_vect=None):
    """

    - si grouping_vect == [], on calcule le taux de remplissage sur la totalité de la matrice
    - si grouping_vect n'est pas vide, c'est un DF indexé comme mat (par des ids d'objets)
    et le taux de remplissage d'une colonne est calculé séparément sur tous les groupes d'objets
    partageant le même identifiant de regroupement

    """

    if grouping_vect is None:
        grouping_vect = []
    if type(mat) is pd.DataFrame:
        cols = mat.columns
    else:
        cols = list(["F" + str(x) for x in range(1, mat.shape[1] + 1)])
        mat = pd.DataFrame(data=mat.values, columns=cols)
    # cols.append ('Avg_rate_by_group')

    if len(grouping_vect) == 0:
        fill_index = [0, "Avg_rate_by_feature"]
        # attribution du groupe 0 à toutes les observations
        grouping_vect = pd.DataFrame(index=mat.index, columns=["Grp"])
        grouping_vect["Grp"] = 0
        # grp_values = {0}

    elif type(grouping_vect) in [pd.Series, pd.DataFrame]:
        grp_values = list(set(grouping_vect[grouping_vect.columns[0]]))
        fill_index = grp_values
        # noinspection PyTypeChecker
        fill_index.append("Avg_rate_by_feature")

    else:
        return

    df_fill_rate = pd.DataFrame(index=fill_index, columns=cols)
    df_fill_rate["Avg_rate_by_group"] = 0.0
    fill_index.remove("Avg_rate_by_feature")

    for idx in fill_index:
        for col in cols:
            subindex = grouping_vect[grouping_vect[grouping_vect.columns[0]] == idx].index
            subindex = mat.index.intersection(subindex)
            colvalues = mat.loc[subindex, col]
            colvalues = colvalues.dropna(inplace=False)
            if len(subindex) > 0:
                df_fill_rate.loc[idx, col] = float(len(colvalues) / len(subindex))

    df_fill_rate.loc["Avg_rate_by_feature", :] = np.nanmean(df_fill_rate, axis=0)
    df_fill_rate.loc[:, "Avg_rate_by_group"] = np.nanmean(df_fill_rate, axis=1)

    if len(grouping_vect) == 0:
        # pas de groupement des observations: un seul taux de remplissage par feature
        # Tri des colonnes par taux décroissant
        df_fill_rate = df_fill_rate.sort_values(by="Avg_rate_by_feature", axis=1, ascending=False, inplace=False)
    else:
        df_fill_rate = df_fill_rate.sort_values(by="Avg_rate_by_feature", axis=1, ascending=False, inplace=False)
        df_fill_rate = df_fill_rate.sort_values(by="Avg_rate_by_group", axis=0, ascending=False, inplace=False)

    return df_fill_rate.astype(float)


def sort_lines_by_fill_rate(mat):
    if type(mat) is pd.DataFrame:
        lines = mat.index
    else:
        lines = list(["I" + str(x) for x in range(1, mat.shape[0] + 1)])
        mat = pd.DataFrame(data=mat.values, index=lines)
    df_fill_rate = pd.DataFrame(index=[0], columns=lines)
    for line in lines:
        linevalues = mat.loc[line]
        linevalues = linevalues.dropna(inplace=False)
        df_fill_rate[line] = float(len(linevalues) / mat.shape[1])

    df_fill_rate = df_fill_rate.sort_values(by=0, axis=1, ascending=False)
    return df_fill_rate.astype(float)


def drop_sparse_lines(mat, minimum_fill_rate=0.5):
    if type(mat) is pd.DataFrame:
        cols = mat.columns
        lines = mat.index
    else:
        cols = list(["F" + str(x) for x in range(1, mat.shape[1] + 1)])
        lines = list(["I" + str(x) for x in range(1, mat.shape[0] + 1)])
    submat = pd.DataFrame(data=mat.values, columns=cols, index=lines)
    # df_lfill_rate = pd.DataFrame(index=[0],
    #                             columns=lines)
    lines_to_drop = []

    for line in lines:
        linevalues = mat.loc[line]
        linevalues = linevalues.dropna(inplace=False)
        if float(len(linevalues) / submat.shape[1]) < minimum_fill_rate:
            lines_to_drop.append(line)

    submat = submat.drop(lines_to_drop, axis=0)

    return submat


def drop_sparse_columns(mat, minimum_fill_rate=0.5) -> (pd.DataFrame, list):
    """

    drop_sparse_columns:
    Eliminates columns from a matrix for which the proportion of filled (not nan) values
    to the number of entries is less than a given minimum.

    Parameters
    ----------

    mat: a bi-dimensional numeric array: float[,] ndarray or DataFrame
    If mat has no index/columns, a DataFrame with automated index and columns is created.

    minimum_fill_rate: the minimum ratio. All columns with a strictly lower fill rate are dropped.

    returns
    -------

    pd.DataFrame, list: A copy of the original matrix (as DataFrame), without the sparse columns and the list of
    dropped columns
    """

    if type(mat) is pd.DataFrame:
        cols = mat.columns
        lines = mat.index
    else:
        cols = list(["F" + str(x) for x in range(1, mat.shape[1] + 1)])
        lines = list(["I" + str(x) for x in range(1, mat.shape[0] + 1)])

    submat = pd.DataFrame(data=mat.values, columns=cols, index=lines)
    # df_cfill_rate = pd.DataFrame(index=[0],
    #                             columns=cols)
    cols_to_drop = []
    for col in cols:
        colvalues = mat[col]
        colvalues = colvalues.dropna(inplace=False)
        if float(len(colvalues) / mat.shape[0]) < minimum_fill_rate:
            cols_to_drop.append(col)

    submat = submat.drop(cols_to_drop, axis=1)

    return submat, cols_to_drop


def drop_zero_columns(mat):
    """

    drop_null_columns:
    Eliminates from a matrix those columns only populated by 0.

    Parameters
    ----------

    :param mat: a bi-dimensional numeric array: float[,] ndarray or DataFrame
    If mat has no index/columns, a DataFrame with automated index and columns is created.

     :returns
    --------

    A copy of the original matrix (as DataFrame), without the null columns
    The list of dropped columns
    """

    if type(mat) is pd.DataFrame:
        cols = mat.columns
        lines = mat.index
    else:
        cols = list(["F" + str(x) for x in range(1, mat.shape[1] + 1)])
        lines = list(["I" + str(x) for x in range(1, mat.shape[0] + 1)])

    submat = pd.DataFrame(data=mat.values, columns=cols, index=lines)
    # df_cfill_rate = pd.DataFrame(index=[0],
    #                             columns=cols)
    cols_to_drop = []
    for col in cols:
        colvalues = mat[col]
        if all([v == 0 for v in colvalues]):
            cols_to_drop.append(col)

    submat = submat.drop(cols_to_drop, axis=1)

    return submat, cols_to_drop


def drop_lines_and_columns(mat, minimum_fill_rate=0.5, droplines=True, dropcols=True, colsfirst=True):
    if droplines and dropcols:
        if colsfirst:
            submat = drop_sparse_columns(mat, minimum_fill_rate)
            submat = drop_sparse_lines(submat, minimum_fill_rate)
        else:
            submat = drop_sparse_lines(mat, minimum_fill_rate)
            submat = drop_sparse_columns(submat, minimum_fill_rate)
    elif dropcols:
        submat = drop_sparse_columns(mat, minimum_fill_rate)
    else:
        submat = drop_sparse_lines(mat, minimum_fill_rate)

    return submat


def drop_features(mat, features_to_drop):
    submat = mat.copy().drop(features_to_drop, axis=1) if features_to_drop is not None else mat.copy()
    return submat


def separate_along_median_by_cols(mat):
    cols_plus = [str(x) + "_+" for x in mat.columns]
    cols_minus = [str(x) + "_-" for x in mat.columns]
    cols_meds = mat.astype(float).apply(np.nanmedian, axis=0)
    mat_plus = pd.DataFrame(index=mat.index, data=mat.values, columns=cols_plus)

    mat_minus = pd.DataFrame(index=mat.index, data=mat.values, columns=cols_minus)
    m = pd.concat([mat_plus, mat_minus], axis=1)
    for col in mat.columns:
        m[str(col) + "_+"] = np.maximum(0.0, mat[col] - cols_meds[col])
        m[str(col) + "_-"] = np.maximum(0.0, cols_meds[col] - mat[col])
    return m, cols_meds


def rebuild_median_separated_matrix(duplicated_mat, med):

    if med is None:
        pass
    [nf] = med.shape
    _, nmf = duplicated_mat.shape
    assert nmf == 2 * nf
    origin_mat = pd.DataFrame(columns=med.index, index=duplicated_mat.index, data=0)
    origin_mat = origin_mat.add(duplicated_mat.iloc[:, 0:nf].values)
    origin_mat = origin_mat.sub(duplicated_mat.iloc[:, nf: 2 * nf].values)
    origin_mat = origin_mat.T.add(med, axis="index").T
    return origin_mat


def plot_fill_rate(mat, bycols=True, grouping_vect=None, path=None):
    if grouping_vect is None:
        grouping_vect = []
    fill_rate = sort_columns_by_fill_rate(mat, grouping_vect) if bycols else sort_lines_by_fill_rate(mat)

    axis = 1 if bycols else 0

    f, ax = plt.subplots(figsize=(10, 5))
    if len(grouping_vect) == 0:
        mean_rate = np.mean(fill_rate, axis=axis)[0]
        plt.bar(
            list(range(len(fill_rate.columns))),
            fill_rate.loc[0, :], width=2.0, label="Mean {0:.1%}".format(mean_rate)
        )
        ax.set_ylabel("Fill rate")
        xlabel = "Features" if bycols else "Stocks"
        ax.set_xlabel(xlabel)
        ax.set_title(xlabel + " ordered by fill rate")
        ax.legend()
    else:
        fill_rate = fill_rate.astype(float)

        _, _ = plt.subplots(figsize=(20, 10))
        _ = sns.heatmap(
            fill_rate,
            annot=False,
            linewidths=0,
            yticklabels=True,
            xticklabels=True,
            cmap="coolwarm",
            cbar_kws={"label": "Fill rate"},
        )
    if path is not None:
        plt.gcf().savefig(path)
    plt.show()


def get_features_correlation_matrix(mat):
    mat_corr = pd.DataFrame(index=mat.columns, columns=mat.columns)
    for irow, row in enumerate(mat_corr.index):
        for icol, col in enumerate(mat_corr.columns):
            if icol == irow:
                mat_corr.iloc[irow, icol] = 1
            elif icol < irow:
                mat_corr.iloc[irow, icol] = mat_corr.iloc[icol, irow]
            else:
                mat_corr.iloc[irow, icol] = np.corrcoef(mat.iloc[:, irow], mat.iloc[:, icol])[0, 1]

    return mat_corr


def plot_values(mat, xticklabels=False, yticklabels=True, path=None):
    if not type(mat) is pd.DataFrame:
        cols = list(["F" + str(x) for x in range(1, mat.shape[1] + 1)])
        lines = list(["I" + str(x) for x in range(1, mat.shape[0] + 1)])
        mat = pd.DataFrame(index=cols, columns=lines, data=mat)
    plt.subplots(figsize=(40, 20))
    sns.heatmap(mat, annot=False, linewidths=0, xticklabels=xticklabels, yticklabels=yticklabels, cmap="bwr")

    if path is not None:
        plt.gcf().savefig(path)
    else:
        plt.show()
    plt.close("all")


def generate_random_correl_matrix(size: int, corr_max: float):
    """

     generate_random_correl_matrix:
     generate a random correlation matrix having an absolute maximum correlation

     Parameters
     ----------

     size: the size of the square matrix

     corr_max: the maximum of absolute value of non-diagonal terms

     :returns:
     the desired correlation matrix as DataFrame
     """

    assert (size > 0)
    mat = np.ndarray([size, size])
    res = pd.DataFrame(mat)
    for col in range(size):
        mat.iloc[col, :] = np.random.uniform(-corr_max, corr_max, size)
        mat.iloc[col, col] = 1
    return res


def generate_orthogonal_vector(v: object) -> pd.DataFrame:
    """

     generate_orthogonal_vector:
     generate a random vector of the same size as a given non null vector v, orthogonal to v

     Parameters
     ----------

     v: the vector

     :returns:
     the random vector orthogonal to v
     """
    if type(v) is list:
        v = pd.DataFrame(v)
    n, _ = v.shape
    assert (n > 1)
    assert (not all([v.iloc[i, 0] == 0 for i in range(n)]))
    w = pd.DataFrame(np.zeros(n))
    if n == 2:
        w.iloc[0, 0] = v.iloc[1, 0]
        w.iloc[1, 0] = -v.iloc[0, 0]
    else:
        istar = np.argmin([1 if v.iloc[i, 0] != 0 else 0 for i in range(n)])
        w.iloc[:, 0] = np.random.uniform(-1, 1, n)
        vw = np.dot(v.T, w)[0, 0]
        w.iloc[istar, 0] = - vw / v.iloc[istar, 0]
    return w


def generate_vectors_bouquet(vpivot, bouquet_size: int, min_corr: float,
                             positive: bool = True,
                             col_shift: int = 0,
                             row_shift: int = 0) -> pd.DataFrame:

    vpivot = pd.DataFrame(vpivot)
    nr, _ = vpivot.shape
    df_bouquet = pd.DataFrame(index=range(row_shift, row_shift+nr),
                              columns=range(col_shift, col_shift+bouquet_size))
    df_bouquet.fillna(value=0.0, inplace=True)
    epsilon = (1 - min_corr)**0.5
    n2v = norm_n2(vpivot)
    for icol in range(bouquet_size):
        w = generate_orthogonal_vector(vpivot)
        n2w = norm_n2(w)
        rescaled_w = epsilon * n2v / n2w * pd.DataFrame(w)
        vals = vpivot.values + rescaled_w.values
        df_bouquet.loc[:, col_shift+icol] = vals
    if positive:
        minval = np.min(np.min(df_bouquet))
        if minval < 0:
            df_bouquet -= minval
    return df_bouquet


def generate_clusters_centroids(ndim: int, max_corr: float) -> pd.DataFrame:

    assert (ndim > 1)
    assert (0 <= max_corr < 1)
    a = ndim / 4.0 * (1 - max_corr / 3)
    b = 1 - max_corr
    c = - max_corr
    delta = b*b - 4*a*c
    epsmax = 0.5 * (- b + delta**0.5) / a
    df_centroids = pd.DataFrame(index=range(ndim), columns=range(ndim))
    df_centroids.fillna(value=0.0, inplace=True)
    for icol in range(ndim):
        df_centroids.iloc[icol, icol] = 1
        df_centroids.iloc[:, icol] += np.random.uniform(0.0, 2*epsmax, ndim)

    return df_centroids


def generate_integer_partition(ntot: int, nbpart: int, nmin: int = 1):

    assert(ntot >= nbpart*nmin)
    rng = np.random.default_rng()
    cuts = np.sort(rng.choice(ntot, nbpart, replace=False))
    res = []
    zum = 0
    defect = 0
    for i, cut in enumerate(cuts):
        diff = cut if i == 0 else cut - cuts[i-1]
        if diff < nmin:
            defect += nmin - diff
            val = nmin
        elif defect == 0:
            val = diff
        elif diff >= nmin + defect:
            val = diff - defect
            defect = 0
        else:
            val = diff
        res.append(val)
        zum += val
    if zum < ntot:
        res[-1] += ntot-zum
    # partsize = ntot / nbpart
    # res = [partsize] * nbpart
    return res


def generate_clusterized_vectors(veclen: int,
                                 nclusters: int,
                                 nbvect: int,
                                 clusterminsize: int,
                                 max_inter_corr: float,
                                 min_intra_corr: float,
                                 random_norms: bool = True) -> pd.DataFrame:

    clustersizes = generate_integer_partition(ntot=nbvect, nbpart=nclusters, nmin=clusterminsize)
    centroids = generate_clusters_centroids(ndim=veclen, max_corr=max_inter_corr)
    df_res = pd.DataFrame(index=range(veclen))
    # factors = np.random.exponential(size=nclusters) if random_norms else np.zeros(shape=nclusters)
    factors = np.random.lognormal(mean=0, sigma=1, size=nclusters)
    row_shift = 0
    col_shift = 0
    for icluster, csize in enumerate(clustersizes):
        centroid = centroids.iloc[:, icluster]
        df_bouquet = generate_vectors_bouquet(vpivot=centroid,
                                              bouquet_size=int(csize),
                                              min_corr=min_intra_corr,
                                              col_shift=col_shift
                                              )
        df_bouquet *= 1 + factors[icluster]
        df_res = pd.concat([df_res, df_bouquet], axis=1)
        col_shift += int(csize)
    return df_res


def build_correspondence_map(df_clusters) -> pd.DataFrame:

    assert len(df_clusters.columns) >= 2
    clusters_1 = [str(x) for x in set(df_clusters.iloc[:, 0])]
    clusters_2 = [str(x) for x in set(df_clusters.iloc[:, 1])]
    df_map = pd.DataFrame(index=clusters_1, columns=clusters_2)
    df_map.fillna(value=0.0, inplace=True)
    for idx in df_clusters.index:
        cl1 = str(df_clusters.loc[idx][0])
        cl2 = str(df_clusters.loc[idx][1])
        df_map.loc[cl1, cl2] += 1

    sum_of_cluster1 = df_map.apply(np.sum, axis=1)
    for idx in df_map.index:
        df_map.loc[idx, :] /= sum_of_cluster1[idx]

    return df_map


def plot_correspondence_map(df_clusters, path=None):

    df_map = build_correspondence_map(df_clusters)
    plot_values(df_map, xticklabels=True, path=path)
    return None


def duplicate_data_wfold(df: pd.DataFrame, window: int) -> pd.DataFrame:
    if window <= 0:
        return pd.DataFrame()
    cols = df.columns
    nidx = len(df.index)
    l_shifted_cols = [[f"{c}_{i}" for i in range(window)] for c in cols]
    shifted_cols = [item for sublist in l_shifted_cols for item in sublist]
    duplicated_df = pd.DataFrame(index=df.index, columns=shifted_cols)
    for irow, idx in enumerate(df.index):
        for icol, col in enumerate(cols):
            for shift in range(window):
                duplicated_df.loc[idx, f"{col}_{shift}"] = df.iloc[irow+shift, icol] if irow+shift < nidx else np.nan
    return duplicated_df




from pathlib import Path

import kfp.dsl as dsl
import kfp
from kubernetes.client.models import V1Toleration
import argparse


# noinspection PyUnresolvedReferences
def function(prog_name, image, memory_limit, cpu):
    container_op = dsl.ContainerOp(
        name=prog_name.stem,
        image=f"eu.gcr.io/second-capsule-253207/{image}:latest",
        command=["python"],
        arguments=[str(prog_name)],
    )
    # Single job specification
    container_op.add_resource_limit("memory", memory_limit)
    container_op.add_resource_limit("cpu", cpu)
    container_op.add_resource_request("memory", memory_limit)
    container_op.add_resource_request("cpu", cpu)
    container_op.add_toleration(
        V1Toleration(key="dedicated", operator="Equal", value="standard-preemptible", effect="NoSchedule")
    )
    return container_op


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Creates the yaml file to launch several iterations of a program on GCP",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument("-p", "--program", type=str, default="", help="Program .py file name")
    parser.add_argument("-i", "--iterations", type=int, default=1, help="Number of times to run the program")
    parser.add_argument(
        "-m", "--max_executions", type=int, default=15, help="Max number of runs that can exist at a time"
    )
    parser.add_argument(
        "-d", "--docker", type=str, default="", help="Docker image name. Default will be the program file name"
    )
    parser.add_argument("-r", "--ram", type=str, default="2G", help="Max ram used by one single run")
    parser.add_argument("-c", "--cpu", type=str, default="1", help="Max number of CPUs used by one single run")
    args = parser.parse_args()

    program = args.program
    iterations = args.iterations
    image_name = args.docker
    parallelism = args.max_executions
    ram = args.ram
    cpus = args.cpu

    program_path = Path(program)

    if not program_path.is_file():
        raise FileNotFoundError(f"Could not find file '{program_path}'")

    if not isinstance(iterations, int):
        raise TypeError("Number of iterations must be an integer")

    if iterations < 0:
        raise ValueError("Number of iterations must be greater than 0")

    if image_name == "":
        image_name = program_path.stem

    @dsl.pipeline(name="function")
    def pipeline():
        loop_args = list(range(iterations))  # run 20 jobs
        with dsl.ParallelFor(loop_args, parallelism=parallelism) as _:  # On max 10 machines at a time
            function(program_path, image_name, ram, cpus)

    # noinspection PyUnresolvedReferences
    kfp.compiler.Compiler().compile(pipeline, str(program_path.with_suffix(".yaml")))



import pandas as pd
import os
import numpy as np
# import math

# from sklearn.decomposition import NMF

# from adlinear import pca
from adlinear import utilities as utl
from adlinear import nmfmodel as nmf
from adlinear import testers as tst
from adlinear import features_selector as fs
from randomgenerators import randomgenerators as rng
# from adlinear import ntfmodel as ntf
from sklearn.svm import SVC, SVR
from sklearn.linear_model import ElasticNet

# import matplotlib
# import matplotlib.pyplot as plt
# import matplotlib.ticker as mtick
# import seaborn as sns
from pathlib import Path

import rulefit
import CoveringAlgorithm.CA as CA
from sklearn.ensemble import RandomForestRegressor

import root_path
import dotenv
dotenv.load_dotenv()


if __name__ == "__main__":

    root_path = root_path.get_root_path()

    rd_path = root_path / os.getenv("rd_subpath")
    sparsity_paper_path = rd_path / os.getenv("sparsity_paper_subpath")

    data_path = sparsity_paper_path / os.getenv("data_subpath")

    temp_data_path = data_path / "temp_data/"
    pictures_path = data_path / "Pics"
    out_data_path = sparsity_paper_path / "results/"
    codalab_data_path = sparsity_paper_path / "codalab"

    if os.getenv("sp_do_brunet", "False").lower() == "true":
        # Récupérer le jeu de données
        df_brunet = pd.read_csv(data_path / os.getenv("brunet_filename"))
        # récupérer le fichier des groupes
        df_brunet_groups = pd.read_csv(data_path / os.getenv("brunet_grp_filename"))
        df_brunet.loc[:, "Group"] = df_brunet_groups.loc[:, "Group"]
        # la colonne d'attribution des groupes
        df_brunet_priors = df_brunet.replace({"Group": {"ALL_B_1": 1,
                                                        "ALL_B_2": 2,
                                                        "ALL_T": 3,
                                                        "AML": 4}})["Group"]
        # Enlever la première colonne d'étiquettes
        df_brunet = df_brunet[df_brunet.columns[2:]]
        # Appliquer la transformation log par rapport au minimum
        df_log_brunet = df_brunet - np.nanmin(df_brunet, axis=0)
        df_log_brunet = np.log(1.0 + df_brunet)

        # Fixer le nombre de composants nmf
        ncomp_brunet = 4
        nmf_brunet = nmf.NmfModel(mat=df_brunet, name="brunet",
                                  ncomp=ncomp_brunet, regularization="components", leverage="robust", max_iter=200)
        nmf_log_brunet = nmf.NmfModel(mat=df_log_brunet, name="log_brunet",
                                      ncomp=ncomp_brunet, regularization="components", leverage="robust", max_iter=200)

#        sparse_res = tst.test_nmf_sparsity(df_brunet, priors=df_brunet_priors, ncomp=ncomp_brunet, nblocks=1,
#                                               sp_step=0.1, sp_min=0, sp_max=1,
#                                               outpath=out_data_path, do_save_result=True, name="brunet")
        sparse_res2 = tst.test_nmf_sparsity(df_log_brunet, priors=df_brunet_priors, ncomp=ncomp_brunet, nblocks=2,
                                                sp_step=0.1, sp_min=0, sp_max=1,
                                                outpath=out_data_path, do_save_result=True, name="log_brunet")
        brunet_sparsity = 0.0
    # Dataprep Arcene
    if os.getenv("sp_do_arcene", "True").lower() == "true":
        # data_path = Path('/home/cgeissler/local_data/Article_sparsity/data/')
        dataset_name = os.getenv("arcene_name")
        df_arcene_X_tr = pd.read_csv(data_path / os.getenv("arcene_train_filename"),
                                     sep=" ",
                                     header=None)
        # récupérer les en-tête de colonnes comme chaines
        df_arcene_X_tr.columns = [str(x) for x in df_arcene_X_tr.columns]
        nf = len(df_arcene_X_tr.columns)
        # remove last column full on nan's
        df_arcene_X_tr = df_arcene_X_tr.iloc[:, 0:nf - 2]

        # lecture des étiquettes d'apprentissage
        df_arcene_y_tr = pd.read_csv(data_path / os.getenv("arcene_train_labels"),
                                     sep=" ",
                                     header=None)
        # lecture des variables de test
        df_arcene_X_te = pd.read_csv(data_path / os.getenv("arcene_test_filename"),
                                     sep=" ",
                                     header=None)
        df_arcene_X_te.columns = [str(x) for x in df_arcene_X_te.columns]
        nf = len(df_arcene_X_te.columns)
        # remove last column full on nan's
        df_arcene_X_te = df_arcene_X_te.iloc[:, 0:nf - 2]

        # lecture des variables de validation
        df_arcene_X_va = pd.read_csv(data_path / os.getenv("arcene_valid_filename"),
                                     sep=" ",
                                     header=None)
        df_arcene_X_va.columns = [str(x) for x in df_arcene_X_va.columns]
        nf = len(df_arcene_X_va.columns)
        # remove last column full on nan's
        df_arcene_X_va = df_arcene_X_va.iloc[:, 0:nf - 2]
        # étiquettes de validation
        df_arcene_y_va = pd.read_csv(data_path / os.getenv("arcene_valid_labels"),
                                     sep=" ",
                                     header=None)

        # Screeplot = dépendance au nombre de composants NMF sur arcene training
        if os.getenv("sp_do_arcene_screeplot", "True").lower() == "true":
            # Tests nmf en parcimonie adaptative et pour différents nombres de composants
            screeplot_res_snmf = tst.test_nmf_ncomp(df_arcene_X_tr, priors=None, ncomp_min=21, ncomp_max=20, nblocks=2,
                                                    sparsity=1.0,
                                                    outpath=out_data_path, do_save_result=True, name=dataset_name)
            screeplot_res_nmf = tst.test_nmf_ncomp(df_arcene_X_tr, priors=None, ncomp_min=2, ncomp_max=17, nblocks=1,
                                                   sparsity=1.0,
                                                   outpath=out_data_path, do_save_result=True, name=dataset_name)
        # Etude d'influence de la parcimonie sur arcene
        if os.getenv("sp_do_arcene_sparsity", "True").lower() == "true":
            # Tests nmf avec 8 composants et différents niveaux de parcimonie
            sparse_res_nmf = tst.test_nmf_sparsity(df_arcene_X_tr, priors=None, ncomp=8, nblocks=1,
                                                   sp_step=0.20, sp_min=0, sp_max=1,
                                                   outpath=out_data_path, do_save_result=True, name=dataset_name)

            sparse_res_ntf = tst.test_nmf_sparsity(df_arcene_X_tr, priors=None, ncomp=9, nblocks=2,
                                                   sp_step=0.20, sp_min=0, sp_max=1,
                                                   outpath=out_data_path, do_save_result=True, name=dataset_name)
        # Influence de la sélection de variables sur le modèle prédictif arcene / svm
        if os.getenv("sp_do_arcene_svm", "True").lower() == "true":
            # Tester la performance d'un modèle
            models_params = [[SVC, ]]
            x_train = df_arcene_X_tr
            x_valid = df_arcene_X_va
            y_train = df_arcene_y_tr
            y_valid = df_arcene_y_va
            x_test = df_arcene_X_te

            for model_params in models_params:
                model = models_params[0]
                params = models_params[1]
                full_model = model.__init__(params)
                full_model.fit(x_train, y_train)
                score_tr_full = full_model.score(x_train, y_train)
                score_va_full = full_model.score(x_valid, y_valid)
                predict_test_full = full_model.predict(x_test)

                arcene_nmf = nmf.NmfModel(mat=x_train, ncomp=8, name=dataset_name, regularization="components")
                arcene_nmf.set_sparsity(1.0)

                ht = arcene_nmf.get_h().T
                _, null_arcene_features = utl.drop_zero_columns(ht)

                df_arcene_X_tr_red = df_arcene_X_tr.drop(null_arcene_features, axis=1) \
                    if null_arcene_features \
                    else df_arcene_X_tr
                df_arcene_X_va_red = df_arcene_X_va.drop(null_arcene_features, axis=1) \
                    if null_arcene_features \
                    else df_arcene_X_va
                df_arcene_X_te_red = df_arcene_X_te.drop(null_arcene_features, axis=1) \
                    if null_arcene_features \
                    else df_arcene_X_te

                arcene_nmf_red = nmf.NmfModel(mat=df_arcene_X_tr_red, ncomp=8, name=dataset_name,
                                              regularization="components")
                fuzzy_features = arcene_nmf_red.get_fuzzy_features(1)
                df_arcene_X_tr_red = df_arcene_X_tr_red.drop(fuzzy_features, axis=1) \
                    if fuzzy_features \
                    else df_arcene_X_tr
                df_arcene_X_va_red = df_arcene_X_va_red.drop(fuzzy_features, axis=1) \
                    if fuzzy_features \
                    else df_arcene_X_va
                df_arcene_X_te_red = df_arcene_X_te_red.drop(fuzzy_features, axis=1) \
                    if fuzzy_features \
                    else df_arcene_X_te

                svc_model_red = SVC()
                svc_model_red.fit(df_arcene_X_tr_red, df_arcene_y_tr)
                score_tr_red = svc_model_red.score(df_arcene_X_tr_red, df_arcene_y_tr)
                score_va_red = svc_model_red.score(df_arcene_X_va_red, df_arcene_y_va)
                predict_test_red = svc_model_red.predict(df_arcene_X_te_red)
                predict_test_red = np.maximum(predict_test_red, 0.0)
                np.savetxt(Path(out_data_path) / f"arcene_test.predict", predict_test_red, fmt="%.0f")
                np.savetxt(Path(out_data_path) / f"arcene_test.feat", df_arcene_X_tr_red.columns, fmt="%s")

                df_svm_scores = pd.DataFrame(index=['Full', 'Reduced'],
                                             columns=['Training', 'Validation'])

                df_svm_scores.loc['Full', 'Training'] = score_tr_full
                df_svm_scores.loc['Reduced', 'Training'] = score_tr_red
                df_svm_scores.loc['Full', 'Validation'] = score_va_full
                df_svm_scores.loc['Reduced', 'Validation'] = score_va_red
                df_svm_scores.to_csv(Path(out_data_path) / f"arcene_svm_scores.csv")

        # Comparaison de méthodes de sélection de variables sur arcene
        if os.getenv("sp_do_arcene_fs", "True").lower() == "true":

            f_sel = fs.FeatureSelector(method="random")
            # Testeur de sélection sur
            selection_tester = fs.FeatureSelectorTester("Arcene_SVC_tester_w_random", predictive_template=None,
                                                        fselectors=[f_sel], dataset_name=dataset_name,
                                                        training_set=df_arcene_X_tr, validation_set=df_arcene_X_va,
                                                        test_set=df_arcene_X_te, training_labels=df_arcene_y_tr,
                                                        validation_labels=df_arcene_y_va, classifying_y=True)
            nb_rand_trials = 100
            selection_tester.set_predictive_model(SVC)
            # Test de RuleFit
            selection_tester.set_predictive_model(rulefit.RuleFit,
                                                  tree_size=4,
                                                  max_rules=500,
                                                  random_state=140,
                                                  rfmode="regress")
            # selection_tester.set_predictive_model(CA.CA)

            for ncomp in range(3, 8):
                non_null_snmf_fsel = fs.FeatureSelector(method="snmf", ncomp=ncomp, sparsity=1.0, nblocks=2)

                non_fuzzy_snmf_fsel = fs.FeatureSelector(name=f"OptimalNonFuzzy_{ncomp}", method="snmf", nblocks=2,
                                                         ncomp=ncomp, sparsity=1.0,
                                                         what_to_drop="fuzzy",
                                                         fuzzy_rank=-1)
                non_null_snmf_fsel = fs.FeatureSelector(name=f"NonNull_{ncomp}", method="snmf", nblocks=2,
                                                        ncomp=ncomp, sparsity=1.0,
                                                        what_to_drop="null")
                # Elimination par sparsity
                selection_tester.set_feature_selectors([non_null_snmf_fsel])
                selection_tester.run()
                selection_tester.run_random_selections(nb_rand_trials)

                # Elimination des features nulles, puis des floues
                # selection_tester.set_feature_selectors([non_null_snmf_fsel, non_fuzzy_snmf_fsel])
                # selection_tester.run()
                # selection_tester.output_current_model(dataset_name=dataset_name, outpath=Path(out_data_path))
                # selection_tester.run_random_selections(nb_rand_trials)

                # Eliminer d'abord les features floues puis les nulles
                # selection_tester.set_feature_selectors([non_fuzzy_snmf_fsel, non_null_snmf_fsel])
                # selection_tester.run()
                # non_fuzzy_nmf_fsel.output_features(outpath=Path(out_data_path))
                # selection_tester.output_current_model(dataset_name=dataset_name, outpath=Path(out_data_path))
                # selection_tester.run_random_selections(nb_rand_trials)

            selection_tester.output_all_results(outpath=Path(out_data_path))
            selection_tester.output_current_model(dataset_name=dataset_name, outpath=Path(out_data_path))

    if os.getenv("sp_do_random_fs", "True").lower() == "true":

        standard_noise = rng.RandomVariable(np.random.normal, 1, 0.05)
        signal_prop = 0.50

        dict1 = {"Variable": standard_noise,
                 "Signal_prop": signal_prop,
                 "Coeff": 0.001,
                 "Bias": 0.2,
                 "Min": -10,
                 "Max": 10
                 }
        dict2 = {"Variable": standard_noise,
                 "Signal_prop": signal_prop,
                 "Coeff": 0.001,
                 "Bias": 0.1,
                 "Min": -10,
                 "Max": 10
                 }
        dict3 = {"Variable": standard_noise,
                 "Signal_prop": signal_prop,
                 "Coeff": 0.001,
                 "Bias": -0.2,
                 "Min": -10,
                 "Max": 10
                 }
        dict4 = {"Variable": standard_noise,
                 "Signal_prop": signal_prop,
                 "Coeff": 0.001,
                 "Bias": -0.4,
                 "Min": -10,
                 "Max": 10
                 }
        noise = {"Variable": standard_noise,
                 "Signal_prop": 0.0,
                 "Bias": 0.0,
                 "Min": -10,
                 "Max": 10
                 }

        var_list = [dict1, dict2, dict3, dict4]
        n_clones = 100
        n_samples = 500
        noise_factor = 1
        my_dls = rng.DependentLocalizedSignals(signal_dist=var_list,
                                               non_overlapping_obs=False,
                                               # random_columns=False,
                                               noise_dist=noise,
                                               cloning_mult=n_clones,
                                               n_crossproducts=0,
                                               n_noisecolumns=n_clones*noise_factor,
                                               nsamples=n_samples,
                                               lbound=0.0,
                                               ubound=100.0)
        my_dls_name = my_dls.__repr__()
        _ = my_dls()
        df_rand_samples = my_dls.get_samples()

        # prendre la somme des variables à signal
        y_rand = df_rand_samples.iloc[:, [i * n_clones for i in range(len(var_list))]].sum(axis=1)

        # rnd_ngroups = len(var_list) + 1
        rand_ngroups = len(var_list)
        df_rand_groups = df_rand_samples.loc[:, "Group"]
        df_rand_mat = df_rand_samples.drop("Group", axis=1)

        x_rand_tr = df_rand_mat.iloc[0: int(n_samples / 2), :]
        y_rand_tr = y_rand.iloc[0: int(n_samples / 2)]
        x_rand_val = df_rand_mat.iloc[int(n_samples / 2):, :]
        y_rand_val = y_rand.iloc[int(n_samples / 2):]
        x_rand_test = df_rand_mat.iloc[int(n_samples / 2):, :]
        y_rand_test = y_rand.iloc[int(n_samples / 2):]

        f_sel = fs.FeatureSelector()
        # Testeur de sélection
        # Comm
        selection_tester = fs.FeatureSelectorTester("Random_Regression_tester", predictive_template=None,
                                                    fselectors=[f_sel], training_set=x_rand_tr,
                                                    validation_set=x_rand_val, test_set=x_rand_test,
                                                    training_labels=y_rand_tr, validation_labels=y_rand_val)

        nb_rand_trials = 100
        # Test de RuleFit
        selection_tester.set_predictive_model(rulefit.RuleFit,
                                              tree_size=4,
                                              max_rules=500,
                                              random_state=140,
                                              model_type="regress")
        # selection_tester.set_predictive_model(rulefit.RuleFit, tree_size=4)

        non_null_nmf_fsel = fs.FeatureSelector(name=f"RandomSet_{rand_ngroups}",
                                               method="snmf", nblocks=2,
                                               ncomp=rand_ngroups, sparsity=1.0,
                                               what_to_drop="null")

        selection_tester.set_feature_selectors([non_null_nmf_fsel])
        selection_tester.run()
        selection_tester.run_random_selections(nb_rand_trials)

        # non_fuzzy_nmf_fsel = fs.FeatureSelector(name=f"RandomSet_{rand_ngroups}",
        #                                         method="snmf", nblocks=2,
        #                                         ncomp=rand_ngroups, sparsity=1.0,
        #                                         what_to_drop="fuzzy",
        #                                         fuzzy_rank=-1)
        #
        # selection_tester.set_feature_selectors([non_fuzzy_nmf_fsel])
        # selection_tester.run()
        # selection_tester.run_random_selections(nb_rand_trials)

        selection_tester.output_all_results(outpath=Path(out_data_path))
        pass

    pass



import functools
from pathlib import Path

import numpy.random
from dask.distributed import Client
from dask import distributed as dst

import pandas as pd
import numpy as np
from transparentpath import Path
import dask.dataframe as dd

client = Client("tcp://127.0.0.1:33869")

dfs_path = Path("/home/cgeissler/Documents/Tests")
ddf = dd.read_parquet(dfs_path)

ddf.loc["2020-01-01":"2020-01-03"].groupby('feature_name').mean().compute()


# In[4]:


ind = (((ddf.groupby('feature_name').count().sum(axis=1) / (ddf.index.nunique() * ddf.shape[1])) > 0.5)).compute()
ind.loc[ind].index.to_list()
ddf = ddf.loc[ddf['feature_name'].isin(ind.loc[ind].index.to_list())].compute()


# In[5]:


a = (ddf.groupby('feature_name').max().max(axis=1)) / 2 + (ddf.groupby('feature_name').min().min(axis=1)) / 2
b = (ddf.groupby('feature_name').max().max(axis=1)) / 2 - (ddf.groupby('feature_name').min().min(axis=1)) / 2
ind = ddf.index
ddf = ddf.set_index('feature_name')
for l in a.index:
    ddf.loc[l] = (ddf.loc[l] - a[l]) / b[l]
ddf = ddf.reset_index()
ddf.index = ind


def collect_features_mean_values(data_path: Path, **kwargs) -> pd.DataFrame:
    # exclusion_list=None, period="", mode: str = "mean")
    """

    collect_features_mean_values: collects files into a single matrix
    Each file represent the time history of a given feature.
    The structure of the file must comply with:
    An increasing series of dates as index;
    Object names as columns

    :Parameters
    ----------

    data_path: a directory path indicating the location of data.
        All files must be located in data_path
    exclusion_list: list of strings to avoid in data_path

    period: time interval used to group the observations.
    If unspecified or '', the whole index will be kept for each feature.
    Otherwise, a time interval.

    mode: 'mean' (default), 'first' or 'last' to determine which value to extract

    :returns:
    a DataFrame containing the features (attributes) as columns, the observations (e.g stocks) as lines
    """

    # total files list
    exclusion_list = kwargs.get("exclusion_list", None)
    measurement_dates = kwargs.get("dates", None)
    mode = kwargs.get("mode", "mean")
    do_save = kwargs.get("save", True)
    out_path = kwargs.get("out_path", "")
    if exclusion_list is None:
        exclusion_list = []
    flist = [
        x
        for x in data_path.glob("*")
        if functools.reduce(lambda x, y: x and y, [str(x).find(excl) < 0 for excl in exclusion_list], True)
    ]
    matrix = pd.DataFrame(index=[], columns=[])
    nmax = 0
    for fl in flist:
        feature = fl.stem
        hdf = pd.read_csv(data_path / fl, index_col=0, parse_dates=True).fillna(method="ffill")
        if measurement_dates != "" and measurement_dates is not None:
            hdf = hdf.loc[measurement_dates, :]
            str_dates = f"Set_{len(hdf.index)}_dates"
        else:
            str_dates = ""
        if mode == "mean":
            hdf = pd.DataFrame(hdf.mean(axis=0))
        elif mode == "first":  # last
            hdf = pd.DataFrame(hdf.iloc[0])
        elif mode == "last":
            hdf = pd.DataFrame(hdf.iloc[-1])
        elif mode == "random":
            irnd = numpy.random.Generator.choice(range(len(hdf.index)), 1)[0]
            hdf = pd.DataFrame(hdf.iloc[irnd])
        hdf.columns = [feature]
        if nmax > 0 and len(hdf) >= nmax + 50:
            print((len(hdf), "stocks before."))
            hdf = hdf.loc[matrix.index, :]
            print((len(hdf), "stocks after."))

        if len(hdf) > nmax:
            nmax = len(hdf)
        matrix = pd.concat([matrix, hdf], axis=1, sort=False)
        if do_save:
            mat_name = f"Features_from_{data_path}_{mode}"
            if str_dates != "":
                mat_name += f"at_{str_dates}__.csv"
            matrix.to_csv(Path(out_path) / mat_name)

    return matrix.dropna(how="all")

if __name__ == "__main__":

    pass



import itertools
import os

import numpy as np
import pandas as pd
from nptyping import NDArray
from typing import Callable, Any, Union, Tuple, Dict
from typing import Dict
from adlinear import utilities as utl
from adlinear import nmfmodel as nmf
import root_path
import dotenv
from transparentpath import TransparentPath
import random
TransparentPath.set_global_fs("gcs", bucket="nmf_experiments_dev", token="cred_gcs.json")
dotenv.load_dotenv()

arange = list((0, int(1e6)))


class RandomNamedFile:

    def __init__(self, dirpath: TransparentPath, root_name: str, suffix: str):
        self.dirpath = dirpath
        self.filepath = None
        number = random.randint(*arange)
        filepath = (dirpath / f"{root_name}_{number}").with_suffix(suffix)
        attempts = 0
        while filepath.is_file():
            if attempts > 99:
                # Failed to create a unique file
                return
            number += 1
            attempts += 1
            if number in arange:
                arange.remove(number)
            number = random.randint(*arange)
            filepath = (dirpath / f"{root_name}_{number}").with_suffix(suffix)

        self.filepath = filepath
        self.filepath.touch()


class RandomVariable:

    def __init__(self, npmethod: Callable = np.random.normal, *args, **kwargs):
        """ Specifies a random variable according to a preset distribution and given parameters.

          REVIEW : chacha

          NOTE :

          Parameters
          ----------
          npmethod: distribution

          Returns
          -------

          """
        if not isinstance(npmethod, Callable):
            if isinstance(npmethod, tuple):
                args = sum((npmethod, args), ())
            else:
                args = sum(((npmethod,), args), ())
            npmethod = np.random.normal
        self._method = npmethod
        self._args = args
        self._kwargs = kwargs

    def __call__(self, size: Union[int, Tuple[int, int]] = 1,
                 ) -> Union[Any, NDArray[Any]]:
        """ Generate instances of a random variable.

          REVIEW : chacha

          NOTE :

          Parameters
          ----------
          size: an integer or a pair of integers

          Returns
          -------
          A dataframe of instances
          """
        res = self._method(*self._args, size=size, **self._kwargs)
        if len(res) == 1:
            return res[0]
        return res

    def get_mean(self):
        m = self._kwargs.get("Mean", None)
        if m is None:
            m = self._args[0] if len(self._args) > 0 else 0
        return m

    def apply_bias(self, **kwargs: object) -> Union[Any, NDArray[Any]]:
        """
        apply_bias:
        Applies an additive bias to a random proportion of a index of realizations
        :param kwargs:
        size: integer, total number of realizations
        bias: float, the value of the bias
        coeff: a multiplicative coefficient applied to the deviation between the variable and its mean
        min: a floor applied to the realizations
        max: a ceiling applied to the realizations
        signal_prop: the proportion of biased realizations as a propotion of total size
        :return:
        A vector of realizations with the specified bias applied to an index of realizations
        """
        size = kwargs.pop('size', None)
        if size is None:
            return None
        elif type(size) is list:
            nrows = size[0]
        else:
            nrows = size
        signal_prop = kwargs.pop('signal_prop', None)
        signal_index = kwargs.pop('signal_index', range(nrows))
        bias = kwargs.pop('bias', None)
        std_coeff = kwargs.pop('coeff', None)
        lbound = kwargs.pop('min', None)
        ubound = kwargs.pop('max', None)
        res = self._method(*self._args, size, **kwargs)

        rndg = np.random.default_rng()
        if signal_prop is not None:
            signal_size = min(int(nrows * signal_prop), len(signal_index))
            biased_indices = rndg.choice(signal_index, signal_size, replace=False)
        else:
            biased_indices = range(nrows)
        if std_coeff is not None and std_coeff != 1.0 and len(biased_indices) > 0:
            m = self.get_mean()
            res[biased_indices] = m + std_coeff * (res[biased_indices] - m)
        if bias is not None and bias != 0 and len(biased_indices) > 0:
            res[biased_indices] += bias

        if lbound is not None:
            res = np.maximum(lbound, res)
        if ubound is not None:
            res = np.minimum(ubound, res)
        return res, biased_indices


class RandomGroup:
    def __init__(self, variable: RandomVariable = RandomVariable()):
        self.variable = variable

    def __call__(self, rows: int = 10, columns: int = 10):
        return self.variable((rows, columns))


class DependentLocalizedSignals:
    def __init__(self, signal_dist: Tuple[Dict], noise_dist: dict, non_overlapping_obs=True, cloning_mult: int = 1,
                 n_crossproducts: int = 0, n_noisecolumns: int = 0, nsamples: int = 100, lbound: float = 0.0,
                 ubound: float = 1.0):

        self._signal_dist = signal_dist
        self._nsignals = len(signal_dist)
        self._noise_dist = noise_dist
        self._cloning_mult = cloning_mult
        self._n_crossproducts = n_crossproducts
        self._n_noisecolumns = n_noisecolumns
        self._non_overlapping_signals = non_overlapping_obs
        # self._random_columns = random_columns
        self._nsamples = nsamples
        self._lbound = lbound
        self._ubound = ubound
        self._ncolumns = len(self._signal_dist) * self._cloning_mult + \
            self._n_crossproducts + self._n_noisecolumns
        self._samples = None
        # for distrib
        pass

    def __repr__(self):
        return f"Random_{self._nsamples}x[{self._nsignals}SIG_INDEP={self._non_overlapping_signals}x" \
               f"{self._cloning_mult}_" \
               f"{self._n_crossproducts}CP_{self._n_noisecolumns}Noise] "

    def _make_frame(self):
        n_sig = len(self._signal_dist)
        cols = []
        sig_cols = ["S" + str(i) + "(" + str(j) + ")" for i in range(n_sig) for j in range(self._cloning_mult)]
        cols += sig_cols
        if self._n_crossproducts > 0:
            prod_cols = ["P" + str(i) for i in range(self._n_crossproducts)]
            cols += prod_cols
        noise_cols = ["N" + str(i) for i in range(self._n_noisecolumns)]
        cols += noise_cols
        cols.append("Group")
        self._samples = pd.DataFrame(index=range(self._nsamples),
                                     columns=cols)
        return self

    def get_samples(self):
        if self._samples is None or utl.count_not_nans(self._samples) == 0:
            self.__call__()
        return self._samples

    def get_sub_frame(self, obs_prop=0.5,
                      feat_prop=0.5):
        pass

    def __call__(self):
        self._make_frame()
        m = self._cloning_mult
        # create signal data
        signal_index = range(self._nsamples)
        n_sig = len(self._signal_dist)
        n_prod = self._n_crossproducts
        rndg = np.random.default_rng(seed=42)
        for isig, sig_dict in enumerate(self._signal_dist):
            rvar: RandomVariable = sig_dict.get("Variable", None)
            signal_prop: float = sig_dict.get('Signal_prop', 0.0)
            bias: float = sig_dict.get('Bias', 0.0)
            coeff: float = sig_dict.get('Coeff', 1.0)
            size = [self._nsamples, self._cloning_mult]
            if rvar is not None:
                rvar_clones, rvar_index = rvar.apply_bias(size=size,
                                                          signal_index=signal_index,
                                                          coeff=coeff,
                                                          min=self._lbound,
                                                          max=self._ubound,
                                                          bias=bias,
                                                          signal_prop=signal_prop)
                cloned_cols = [i for i in range(isig * m, (isig + 1) * m)]
                self._samples.iloc[:, cloned_cols] = rvar_clones
                self._samples.loc[rvar_index, "Group"] = isig + 1
                if self._non_overlapping_signals:
                    signal_index = list(set(signal_index).difference(set(rvar_index)))
        # create cross-products
        if n_prod > 0:
            pairs = [i for i in filter(lambda x: x[0] < x[1], itertools.product(range(n_sig * m), range(n_sig * m)))]
            # select random pairs
            rnd_pairs = rndg.choice(pairs, n_prod, replace=False)

            for ipair, pair in enumerate(rnd_pairs):
                col = f"CP{pair[0]}x{pair[1]}"
                self._samples.rename(columns={f"P{ipair}": col}, inplace=True)
                try:
                    self._samples.loc[:, col] = self._samples.iloc[:, pair[0]] * self._samples.iloc[:, pair[1]]
                except ValueError:
                    pass
            pass

        # create noise data
        nvar: RandomVariable = self._noise_dist.get("Variable", None)
        signal_prop = 0.0
        bias = 0.0
        size = [self._nsamples, self._n_noisecolumns]
        noise_col0 = len(self._signal_dist) * self._cloning_mult + self._n_crossproducts

        if nvar is not None and self._n_noisecolumns > 0:
            nvar_clones, _ = nvar.apply_bias(size=size,
                                             min=self._lbound,
                                             max=self._ubound,
                                             bias=bias,
                                             signal_prop=signal_prop)
            self._samples.iloc[:, noise_col0: noise_col0 + self._n_noisecolumns] = nvar_clones

        self._samples.loc[:, "Group"].fillna(value=0.0, inplace=True)
        return self


def generate_nmf_reconstruction(n_comp: int,
                                min_intra_corr: float,
                                max_inter_corr: float,
                                h_avg_clust_size: int,
                                w_avg_clust_size: int,
                                epsilon: float,
                                random_norms: bool = True):

    cl_vects_h = utl.generate_clusterized_vectors(veclen=n_comp,
                                                  nclusters=n_comp,
                                                  nbvect=n_comp * h_avg_clust_size,
                                                  clusterminsize=int(h_avg_clust_size/2),
                                                  max_inter_corr=max_inter_corr,
                                                  min_intra_corr=min_intra_corr,
                                                  random_norms=random_norms)
    cl_vects_w = utl.generate_clusterized_vectors(veclen=n_comp,
                                                  nclusters=n_comp,
                                                  nbvect=n_comp * w_avg_clust_size,
                                                  clusterminsize=int(w_avg_clust_size/2),
                                                  max_inter_corr=max_inter_corr,
                                                  min_intra_corr=min_intra_corr,
                                                  random_norms=random_norms)
    cl_vects_w = cl_vects_w.T
    cl_vects_m = cl_vects_w @ cl_vects_h
    noise = np.random.normal(loc=0.0, scale=epsilon, size=cl_vects_m.shape)
    cl_vects_m *= 1 + noise
    cl_vects_m = np.maximum(cl_vects_m, 0)
    return cl_vects_m


if __name__ == "__main__":
    # localisation du modèle GAN
    root_path = root_path.get_root_path()

    rd_path = root_path / os.getenv("rd_subpath")
    imputation_paper_path = rd_path / os.getenv("imputation_paper_subpath")

    data_path = imputation_paper_path / os.getenv("data_subpath")
    res_path = imputation_paper_path / os.getenv("results_subpath")

    train_f_name = f"wfold_scree_plots.csv"

    if os.path.isfile(res_path / "random_nmf" / train_f_name):
        df_mini_scree_plots = pd.read_csv(res_path / "random_nmf" / train_f_name)
    else:
        df_mini_scree_plots = pd.DataFrame(index=[], columns=[])

    # train_f_name = "foo_scree.csv"
    rand_norms = True
    for itrial in range(200):
        nb_clusters = np.random.randint(low=5, high=40)
        min_corr = np.random.uniform(low=0.75, high=0.95)
        max_corr = np.random.uniform(low=0.05, high=0.25)
        h_avg_csize = np.random.randint(low=10, high=20)
        w_avg_csize = np.random.randint(low=10, high=100)
        eps = np.random.uniform(low=0.0, high=0.25)
        generated_M = generate_nmf_reconstruction(n_comp=nb_clusters,
                                                  min_intra_corr=min_corr,
                                                  max_inter_corr=max_corr,
                                                  h_avg_clust_size=h_avg_csize,
                                                  w_avg_clust_size=w_avg_csize,
                                                  epsilon=eps,
                                                  random_norms=rand_norms)
        rnstr = "RandNorms" if rand_norms else "ConstNorms"
        generated_M.to_csv(res_path / "random_nmf" / f"M_t{itrial}_nc{nb_clusters}_corrmin{round(min_corr,2)}_"
                                       f"corrmax{round(max_corr,2)}_noise{round(eps,2)}_{rnstr}.csv",
                           float_format="%.4f")
        df_scree_plot = nmf.generate_scree_plot(generated_M, ncmin=3, ncmax=45)
        df_mini_scree_plots = nmf.add_miniscree_plots_from_scree_plot(df_mini_scree_plots, df_scree_plot, nb_clusters)
        df_mini_scree_plots.to_csv(res_path / "random_nmf" / train_f_name)

    do_read_mat = False
    if do_read_mat:
        mat_name = "M_nc36_corrmin0.92_corrmax0.13_noise0.11"
        M = pd.read_csv(res_path / "random_nmf" / f"{mat_name}.csv")
        df_scree_plot = nmf.generate_scree_plot(M, ncmin=25, ncmax=40)
        df_scree_plot.to_csv(res_path / "random_nmf" / f"{mat_name}_screeplot.csv")

    pass



import numpy as np
import pandas as pd

from sklearn.cluster import KMeans
from . import utilities as utl
from . import nmfmodel as nmf
from . import ntfmodel as ntf


class Clusterizer:

    def __init__(self,
                 method: str,
                 **kwargs):

        """

          Clusterizer.__init__:

          Parameters
          ----------

          method: a string in
          {
          "set_groups",     # user-defined clustering
          "kmeans",         # applies kmeans (given #groups) to the provided datafame
          "nmf.kmeans",     # applies kmeans to the W matrix from NMF factorization (given #components)
          "nmf.leverage",   # uses leverage-based clustering to the W matrix from NMF factorization (given #components)
          "snmf.kmeans",    # applies kmeans to the W matrix from SNMF factorization (given #components)
          "snmf.leverage"   # uses leverage-based clustering to the W matrix from SNMF factorization (given #components)
          }

          :returns:
          self
          """

        assert method in {"set_groups", "kmeans", "nmf.kmeans", "nmf.leverage", "snmf.kmeans", "snmf.leverage"}
        self._method = method
        self._kwargs = kwargs
        pass

    def __repr__(self):
        name = self._kwargs.get("name", "")
        kwargs = self._kwargs if name == "" else name
        return f"Clust_{self._method}_{kwargs}"

    def get_nbgroups(self):
        return self._kwargs.get("nb_groups", 2)

    def get_nbgroups(self):
        return int(self._kwargs.get("nb_groups", 2))

    def set_nbgroups(self, ng):
        self._kwargs.update(ng, "nb_groups")

    def apply(self,
              mat):
        """
          Clusterizer.apply
          Applies the object's method and parameters to a given matrix

          Parameters
          ----------
          mat: A data matrix

          :returns:
          a one-column dataframe with cluster attributions
        """
        kwargs = self._kwargs
        mat = pd.DataFrame(mat)
        method = self._method.lower()
        ngroups = kwargs.get("nb_groups", 2)
        groups = kwargs.get("groups", None)
        ncomp = kwargs.get("ncomp", ngroups)
        name = kwargs.get("name", "Cl_")
        if method == "set_groups":
            pd.testing.assert_index_equal(groups.index, mat.index)
            res = groups
        elif method == "kmeans":
            km = KMeans(n_clusters=ngroups)
            res = km.fit_predict(mat)
        elif method == "nmf.leverage":
            model = nmf.NmfModel(mat=mat, name=name, ncomp=ngroups,
                                 regularization="components", leverage="robust", max_iter=200)
            res = model.get_obs_leverage_clusters()
        elif method == "nmf.kmeans":
            model = nmf.NmfModel(mat=mat, name=name, ncomp=ncomp,
                                 regularization="components", leverage="robust", max_iter=200)
            res = model.get_obs_kmeans_clusters(ngroups=ngroups)
        elif method == "snmf.leverage":
            model = ntf.NtfModel(mat=mat, name=name, ncomp=ngroups, nblocks=2,
                                 regularization="components", leverage="robust", max_iter=200)
            res = model.get_obs_leverage_clusters()
        elif method == "snmf.kmeans":
            model = ntf.NtfModel(mat=mat, name=name, ncomp=ngroups, nblocks=2,
                                 regularization="components", leverage="robust", max_iter=200)
            res = model.get_obs_kmeans_clusters()
        else:
            res = None

        return None if res is None else pd.DataFrame(res)


def clusterizer_metrics(clst: Clusterizer, mat, ng):
    if clst is None:
        return None
    else:
        clst.set_nbgroups(ng)
        clusters = clst.apply(mat)
        return utl.clustering_caha_score(mat, clusters)


def clusterizer_optimize(clst: Clusterizer, mat, ngmin=2, ngmax=20):
    # Détermine le nb de composants optimal pour une clusterisation
    ngmax = max(ngmin, ngmax)
    df_metrics = pd.DataFrame(index=range(ngmin, ngmax+1), columns=["Score"])
    if clst is None:
        return None
    else:
        for ng in df_metrics.index:
            df_metrics.loc[ng, "Score"] = clusterizer_metrics(clst, mat, ng)
    score_opt = np.argmax(df_metrics["Score"])
    return score_opt, df_metrics



import socket
from pathlib import Path
import os


def get_root_path():
    # Renvoie suivant la config le chemin parent de research_and_development
    host = socket.gethostname()
    force_local = bool(os.getenv("force_local", "False").lower())
    under_ubuntu = (os.environ['GDMSESSION'] == 'ubuntu')
    if not force_local:
        nas_connected = Path(r"/media/SERVEUR/production/research_and_development/").is_dir() if under_ubuntu else \
            Path(r"/davs://advestis.synology.me:5006/").isdir()
    else:
        nas_connected = False

    if host in ["ADLAPTOPCG", "LAPTOP-CG"]:
        # laptop Ubuntu session
        # local = not Path("Y:/").is_dir()
        if not nas_connected:
            rootpath = (Path(r"/home/cgeissler/local_data") if under_ubuntu
                        else Path(r"C:/Users/Christophe Geissler/"))
        else:
            rootpath = Path(r"/media/SERVEUR/production/research_and_development/") if under_ubuntu else \
                       Path(r"/davs://advestis.synology.me:5006/")
    elif host == "cgeissler-portable":
        # Session laptop sous WIndows
        # local = False
        rootpath = Path(r"/davs://advestis.synology.me:5006/") if nas_connected \
            else Path(r"C:/Users/Christophe Geissler/")
        # rootpath = rootpath / r"research_and_development/Reduction_de_Dimension_(NMTF)/Article_PF_clustering/"
    else:
        # session machine fixe Ubuntu
        # local = False
        rootpath = Path(r"/media/SERVEUR/production/") if nas_connected else Path(r"/home/cgeissler/local_data/")
    return rootpath





import pandas as pd
from typing import Union
from typing import List, Iterable, Optional
from typing import Tuple
import numpy as np
import math

from sklearn.decomposition import NMF

from . import pca as pca
from . import utilities as utl
from . import nmfmodel as nmf
from . import ntfmodel as ntf
from . import imputer as imp

# import matplotlib
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick
import seaborn as sns
from pathlib import Path

NMF_MODEL = Union[nmf.NmfModel, ntf.NtfModel]


def compare_full_sparse_nmf_model(m: pd.DataFrame, ncomp: int, nblocks: int = 1, sparsity: float = 0.0,
                                  nfills: int = 0, drop_unrelevant: bool = False, do_plot: bool = False,
                                  pictures_path: str = "") -> Tuple[NMF_MODEL, NMF_MODEL]:
    """

    :param m:
    :param ncomp:
    :param nblocks:
    :param sparsity:
    :param nfills:
    :param drop_unrelevant:
    :param do_plot:
    :param pictures_path:
    :return:
    """

    if nblocks == 1:
        full_model = nmf.NmfModel(mat=m, ncomp=ncomp, regularization="components", leverage="robust", max_iter=200)
    else:
        m_dup, m_med = utl.separate_along_median_by_cols(m)
        full_model = ntf.NtfModel(
            mat=m_dup,
            ncomp=ncomp,
            baseline=m_med,
            nblocks=nblocks,
            regularization="components",
            leverage="robust",
            max_iter=200,
        )

    full_model.set_sparsity(sparsity)
    full_model.test_model(nfill_iters=nfills)
    # nc0 = sparse_model.ncomp
    # # Heatmap NMF
    full_model.sort_features_by_cluster()
    full_model.sort_observations_by_cluster()
    nh = full_model.get_effective_dimension()
    if do_plot:
        full_model.plot_data_heatmap(pictures_path / "Full_nb{}_nc{}_sp{}_nh{}.pdf".format(nblocks, ncomp,
                                                                                           sparsity, nh))

    restricted_model = None
    if sparsity > 0:
        # recalculer le modèle sans sparsity après élimination des colonnes non retenues
        # dans le modèle précédent avec sparsity
        ht = full_model.get_h().T
        h_dense, null_features = utl.drop_zero_columns(ht)
        m_filled_dense = m.drop(null_features, axis=1) if drop_unrelevant else m
        if nblocks == 1:
            restricted_model = nmf.NmfModel(mat=m_filled_dense, ncomp=ncomp, leverage="robust", max_iter=200)
        else:
            m_dup_dense, m_med_dense = utl.separate_along_median_by_cols(m_filled_dense)
            restricted_model = ntf.NtfModel(
                mat=m_dup_dense, ncomp=ncomp, baseline=m_med_dense, nblocks=2, leverage="robust", max_iter=200
            )

        restricted_model.test_model(nfill_iters=nfills)
        restricted_model.sort_features_by_cluster()
        restricted_model.sort_observations_by_cluster()
        nh = restricted_model.get_effective_dimension()
        if do_plot:
            restricted_model.plot_data_heatmap(pictures_path / "restricted_nb{}_nc{}_sp{}_nh{}.pdf".format(nblocks,
                                                                                                           ncomp,
                                                                                                           sparsity,
                                                                                                           nh))

    return full_model, restricted_model


def test_nmf_sparsity(m, ncomp, name="", priors=None, nblocks=1, sp_step=0.1, sp_min=0, sp_max=1,
                      do_save_by_sp_level=False,
                      outpath="", do_save_result=True):
    """
    :param do_save_result:

    """
    assert sp_step > 0
    nstep = int(1.0 / sp_step)
    df_res = pd.DataFrame(
        index=[round(i * sp_step, 2) for i in range(nstep)],
        columns=[],
    )

    sparsity = sp_min
    isp = 0
    modelstr = "ntf" if nblocks > 1 else "nmf"
    modelstr += "_" + name
    if nblocks == 1:
        model = nmf.NmfModel(mat=m, name=f"{name}_{ncomp}_nmf", ncomp=ncomp, regularization="components",
                             leverage="robust", max_iter=200)
    else:
        m_dup, m_med = utl.separate_along_median_by_cols(m)
        model = ntf.NtfModel(
            mat=m_dup,
            name=f"{name}_{ncomp}_snmf",
            ncomp=ncomp,
            baseline=m_med,
            nblocks=nblocks,
            regularization="components",
            leverage="robust",
            max_iter=200)

    while sparsity <= sp_max:
        model.set_sparsity(sparsity)
        model.test_model()
        ht = model.get_h().T
        _, null_features = utl.drop_zero_columns(ht)
        _, null_components = utl.drop_zero_columns(model.get_h())
        m_reduced = m.drop(null_features, axis=1) if null_features else m
        if null_features is None:
            reduced_model = model.copy()
        else:
            if nblocks == 1:
                reduced_model = nmf.NmfModel(mat=m_reduced, name=name, ncomp=ncomp, regularization="components",
                                             leverage="robust",
                                             max_iter=200)
            else:
                m_red_dup, m_red_med = utl.separate_along_median_by_cols(m_reduced)
                reduced_model = ntf.NtfModel(
                    mat=m_red_dup,
                    ncomp=ncomp,
                    baseline=m_red_med,
                    nblocks=nblocks,
                    regularization="components",
                    leverage="robust",
                    max_iter=200)
        reduced_model.set_sparsity(0.0)
        hcomp, hspars = reduced_model.get_h_compression()
        df_res.loc[sparsity, "NComponents"] = ncomp
        df_res.loc[sparsity, "Error_unreduced"] = model.get_precision()
        df_res.loc[sparsity, "Error"] = err = reduced_model.get_precision()
        df_res.loc[sparsity, "H_compression"] = hcomp
        df_res.loc[sparsity, "Volume"] = vol = reduced_model.get_volume()
        df_res.loc[sparsity, "Err/Vol"] = np.nan if vol == 0.0 else err/vol
        df_res.loc[sparsity, "H_sparsity"] = hspars
        df_res.loc[sparsity, "Null_features"] = len(null_features)
        df_res.loc[sparsity, "Null_components"] = len(null_components)
        if nblocks == 1:
            df_res.loc[sparsity, "Relative_entropy"] = reduced_model.get_avg_diag_entropies(nbins=3).iloc[0, 0]
        df_res.loc[sparsity, "Misclassifieds"], _ = reduced_model.get_nb_misclassified_obs(priors) if priors is not None \
            else np.nan, np.nan

        feat_compo = model.get_features_by_decreasing_loading(min_contribution=0.2)
        if do_save_by_sp_level:
            feat_compo.to_csv(Path(outpath) / f"{modelstr}{ncomp}_sp{sparsity}_feat_compo.csv")
            model.get_h().to_csv(Path(outpath) / f"{modelstr}{ncomp}_sp{sparsity}_H.csv")
        isp += 1
        sparsity = round(sparsity + sp_step, 2)
        if sparsity == 1:
            m_reduced.to_csv(Path(outpath) / f"{modelstr}_c{ncomp}_hhi_reduced_data.csv")

    if do_save_result:
        df_res.to_csv(Path(outpath) / f"{modelstr}_c{ncomp}_test_sparsity.csv")

    return df_res


def test_nmf_ncomp(m, ncomp_min=2, ncomp_max=10, name="", priors=None, nblocks=1, sparsity=1.0,
                   outpath="", do_save_result=True):
    """
    :param do_save_result:

    """
    df_res = pd.DataFrame(
        index=[n for n in range(ncomp_min, ncomp_max)],
        columns=["NComponents", "Error_unreduced", "Error", "Volume", "H_compression", "H_sparsity",
                 "Null_features", "Null_components", "Relative_entropy", "Misclassifieds"],
    )

    isp = 0
    modelstr = "ntf" if nblocks > 1 else "nmf"
    modelstr += "_" + name
    if nblocks == 1:
        model = nmf.NmfModel(mat=m, name=f"{name}_nmf", ncomp=ncomp_min, regularization="components",
                             leverage="robust", max_iter=200)
    else:
        m_dup, m_med = utl.separate_along_median_by_cols(m)
        model = ntf.NtfModel(
            mat=m_dup,
            name=f"{name}_snmf",
            ncomp=ncomp_min,
            baseline=m_med,
            nblocks=nblocks,
            regularization="components",
            leverage="robust",
            max_iter=200)

    ncomp = ncomp_min

    while ncomp <= ncomp_max:
        model.set_ncomp(ncomp)
        # attention, refixer la sparsity après le changement de nombre de composants
        model.set_sparsity(sparsity)
        model.test_model()
        ht = model.get_h().T
        _, null_features = utl.drop_zero_columns(ht)
        _, null_components = utl.drop_zero_columns(model.get_h())
        m_reduced = m.drop(null_features, axis=1) if null_features else m
        if null_features is None:
            reduced_model = model.copy()
        else:
            if nblocks == 1:
                reduced_model = nmf.NmfModel(mat=m_reduced, name=name, ncomp=ncomp, regularization="components",
                                             leverage="robust",
                                             max_iter=200)
            else:
                m_red_dup, m_red_med = utl.separate_along_median_by_cols(m_reduced)
                reduced_model = ntf.NtfModel(
                    mat=m_red_dup,
                    ncomp=ncomp,
                    baseline=m_red_med,
                    nblocks=nblocks,
                    regularization="components",
                    leverage="robust",
                    max_iter=200)
        reduced_model.set_sparsity(0.0)
        # hcomp, hspars = reduced_model.get_h_compression()
        df_res.loc[ncomp, "NComponents"] = ncomp
        df_res.loc[ncomp, "Error_unreduced"] = model.get_precision()
        err = df_res.loc[ncomp, "Error"] = reduced_model.get_precision()
        vol = df_res.loc[ncomp, "Volume"] = reduced_model.get_volume()
        df_res.loc[ncomp, "Err/Vol"] = np.nan if vol == 0.0 else err / vol
        df_res.loc[ncomp, "Null_features"] = len(null_features)
        ncomp += 1

    if len(df_res.index) > 0 and do_save_result:
        df_res.to_csv(Path(outpath) / f"{modelstr}_test_ncomp.csv")

    return df_res


def test_missing_proportion(m, ncomp, name, imp_method="mean", grp_priors=None,
                            p_min=0.0, p_step=0.05, p_max=1.0,
                            n_trials=100, do_save_result=True, outpath=""):

    assert p_step > 0
    nstep = max(1, int((p_max - p_min) / p_step))
    df_res = pd.DataFrame(
        index=[round(i * p_step, 2) for i in range(nstep)],
        columns=["NComponents", "Error", "Volume", "Misclassifieds"],
    )
    imp_mean = imp.Imputer("mean", params={})
    imp_nmf_proxy = imp.Imputer("nmf.proxy", params={"ncomp": ncomp, "nfill_iters": 1})
    imp_nmf_fills = imp.Imputer("nmf.proxy", params={"ncomp": ncomp, "nfill_iters": 5})
    for i in range(nstep):
        p = p_min + round(i * p_step, 2)
        if p <= p_max:
            pstr = f"{round(p * 100, 1)} %"
            n_trials = max(n_trials, 1)
            df_trials = pd.DataFrame(index=range(n_trials),
                                     columns=["Error", "Volume", "Misclassifieds"])
            for itrial in range(n_trials):
                m_cens = utl.censor_data(m, p_min + i * p_step, inplace=False)
                filled_m = imp_mean.apply(m, name)
                # model = nmf.NmfModel(mat=m_cens, name=f"{name}_{pstr}_empty", ncomp=ncomp,
                #                     regularization="components", leverage="robust", max_iter=200)
                # df_trials.loc[itrial, "Error"] = imp_mean.get_precision()
                # df_trials.loc[itrial, "Volume"] = imp_mean.get_volume()
                df_trials.loc[itrial, "Misclassifieds"], _ = imp_mean.get_nb_misclassifieds(grp_priors) \
                    if grp_priors is not None \
                    else np.nan

        df_res.loc[p, "NComponents"] = ncomp
        # df_res.loc[p, "MeanError"] = df_trials["Error"].mean()
        # df_res.loc[p, "MaxError"] = df_trials["Error"].max()
        # df_res.loc[p, "MeanVolume"] = df_trials["Volume"].mean()
        # df_res.loc[p, "MinVolume"] = df_trials["Volume"].min()
        df_res.loc[p, "MeanMisClass"] = df_trials["Misclassifieds"].mean()
        df_res.loc[p, "MaxMisClass"] = df_trials["Misclassifieds"].max()

    if do_save_result:
        df_res.to_csv(Path(outpath) / f"{name}_c{ncomp}_test_missing_values.csv")
    return df_res

    pass




import pandas as pd
import numpy as np

from sklearn.decomposition import NMF

import adlinear.data_collector
from adlinear import pca as pca
from adlinear import utilities as utl
from adlinear import nmfmodel as nmf
# from adlinear import ntfmodel as ntf
from adlinear import testers as tst

# import matplotlib
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick
import seaborn as sns
# from pathlib import Path

import root_path

# import bokeh
# from bokeh.plotting import figure, output_file, show
sns.set()


def load_sectors():
    df_sect = pd.read_csv(sectors_path / "Sectors.csv", index_col=0)
    df_w_sect = pd.read_csv(sectors_path / "weights_by_sectors.csv", index_col=0)
    return df_sect, df_w_sect


def load_scores(fname):
    if (temp_data_path / fname).is_file():
        mat = pd.read_csv(temp_data_path / fname, index_col=0)
    else:
        mat = adlinear.data_collector.collect_features_mean_values(
            primary_data_path, exclusion_list=["sector_average", "controversy", "Severity", "conflict", "ecofi"]
        )
        mat.to_csv(temp_data_path / fname)
    return mat


def plot_errors_along_fill_iteration(mat, nc, fills):
    nmf_model = nmf.NmfModel(mat=mat, ncomp=nc, regularization="components")
    nmf_model.set_sparsity(0.5)
    nmf_model.set_n_bootstrap(10)
    nmf_model.test_model(nfill_iters=fills)
    errors = pd.Series(nmf_model.errors)

    fig, ax = plt.subplots(figsize=(8, 4))
    # ax.yaxis.set_major_locator(mtick.MultipleLocator(1.00))
    ax.yaxis.set_minor_locator(mtick.MultipleLocator(1))
    ax.yaxis.set_major_formatter(mtick.PercentFormatter())
    ax.set_ylabel("Error")
    xlabel = "#Trial"

    xa = errors.index
    # noinspection PyTypeChecker
    ax.plot(xa, errors * 100, color="blue", linewidth=4, linestyle="-")

    ax.set_xlabel(xlabel)
    ax.set_title("Error after iterated refills")
    ax.legend()
    plt.show()
    plt.gcf().savefig(pictures_path / "error_along_lines.pdf")


def plot_matrix_fill_rates_by_sector(mat, df_sectors_):
    df_fill_rate = utl.sort_columns_by_fill_rate(mat, df_sectors_).astype(float)
    df_fill_rate["M"] = np.nanmean(df_fill_rate, axis=1)
    df_fill_rate = df_fill_rate.sort_values(by="M", ascending=False, inplace=False)
    df_fill_rate["M"].to_csv(temp_data_path / "fill_rate_by_sector.csv")
    plt.subplots(figsize=(20, 10))
    # plt.xticks(rotation=45)
    ax = sns.heatmap(
        df_fill_rate.drop(columns=["Avg_rate_by_group"]),
        annot=False,
        linewidths=0,
        yticklabels=True,
        xticklabels=True,
        cmap="Blues",
        cbar_kws={"label": "Fill rate"},
    )
    ax.set_xticklabels(ax.get_xticklabels(), rotation=75)
    plt.gcf().savefig(pictures_path / "fill_rates_by_sector.pdf")
    plt.show()


def plot_decreasing_h_values(nmfmodel):
    fig, ax = plt.subplots(figsize=(8, 4))
    # ax.yaxis.set_major_locator(mtick.MultipleLocator(1.00))
    ax.yaxis.set_minor_locator(mtick.MultipleLocator(1))
    ax.yaxis.set_major_formatter(mtick.PercentFormatter())
    ax.set_ylabel("H loading")
    xlabel = ""
    s_h = nmfmodel.get_sorted_h()

    xa = s_h.index

    ax.plot(xa, s_h.iloc[:, 0], color="blue", linewidth=4, linestyle="-")
    # noinspection PyProtectedMember
    for i in range(1, nmfmodel._ncomp):
        ax.plot(xa, s_h.iloc[:, i])

    ax.set_xlabel(xlabel)
    ax.set_title("Sorted H loadings")
    ax.legend(s_h.columns, loc="upper right")
    plt.gcf().savefig(pictures_path / "decreasing_h_values.pdf")
    plt.show()


def compare_nmf_ntf_by_entropy(mat, keep_restricted, drop_unr, nbmin, nbmax, ncmin, ncmax, sp):
    df_delta_entropy = pd.DataFrame(index=range(ncmin, ncmax + 1), columns=range(nbmin, nbmax + 1))
    df_avg_diag_nmf_entropies = None
    df_avg_diag_sntf_entropies = None
    for ncomp_ in range(ncmin, ncmax + 1):
        nmf_model0, nmf_model1 = tst.compare_full_sparse_nmf_model(m=mat, ncomp=ncomp_, sparsity=sp,
                                                                   drop_unrelevant=drop_unr)
        ntf_model0, ntf_model1 = tst.compare_full_sparse_nmf_model(m=mat, ncomp=ncomp_, nblocks=2, sparsity=sp,
                                                                   drop_unrelevant=drop_unr)
        nmf_model = nmf_model1 if keep_restricted and drop_unr else nmf_model0
        ntf_model = ntf_model1 if keep_restricted and drop_unr else ntf_model0
        if nmf_model is not None and ntf_model is not None:
            for nb in range(nbmin, nbmax + 1):
                df_nmf_entropies, _ = nmf_model.get_entropies_by_cluster()
                df_avg_diag_nmf_entropies = nmf_model.get_avg_diag_entropies(nb)
                df_sntf_entropies, _ = ntf_model.get_entropies_by_cluster()
                df_avg_diag_sntf_entropies = ntf_model.get_avg_diag_entropies(nb)
                df_delta_entropy.loc[ncomp_, nb] = df_avg_diag_sntf_entropies.iloc[0, 0] - \
                                                    df_avg_diag_nmf_entropies.iloc[0, 0]

    if nbmax == nbmin and ncmax == ncmin:
        return df_delta_entropy, df_avg_diag_nmf_entropies, df_avg_diag_sntf_entropies
    else:
        return df_delta_entropy, None, None


def plot_correspondence(model, df_sectors_, path=None):

    df_obs_clusters = model.get_obs_leverage_clusters()
    common_idx = set(df_sectors_.index).intersection(set(df_obs_clusters.index))
    df_sect_clust = pd.DataFrame(index=common_idx, columns=["Sector", "NMF cluster"])
    df_sect_clust.loc[common_idx, "Sector"] = df_sectors_.loc[common_idx, df_sectors_.columns[0]]
    df_sect_clust.loc[common_idx, "NMF cluster"] = df_obs_clusters.loc[common_idx, df_obs_clusters.columns[0]]
    # df_corr = utl.build_correspondence_map(df_sect_clust)
    utl.plot_correspondence_map(df_sect_clust, path)


if __name__ == "__main__":

    # Data localization
    rootpath = root_path.get_root_path()
    rootpath = rootpath / r"research_and_development/"
    rootpath = rootpath / r"Reduction_de_Dimension_(NMTF)/Article_PF_clustering/"

    sectors_path = rootpath / "Base/"
    primary_data_path = rootpath / "Base/scors/"
    temp_data_path = rootpath / "temp_data/"
    pictures_path = rootpath / "Pics"
    out_data_path = rootpath / "results/"

    features_file_name = "Raw_features_no28J.csv"
    # Instantiating global variables
    df_sectors, df_w_sectors = load_sectors()

    # test_pca()
    M_unfilled = load_scores(features_file_name)
    M_unfilled.to_csv(temp_data_path / f"unfilled_{features_file_name}")
    M_filled = utl.fill_missing(M_unfilled, method="median")
    M_filled = utl.shorten_features_names(M_filled)
    M_filled.to_csv(temp_data_path / f"trimmed_and_filled_{features_file_name}")

    M_feat_corr = utl.get_features_correlation_matrix(M_filled)
    entropies = [utl.normalized_entropy(M_filled, nbins) for nbins in range(2, 20)]
    ncomp = 6

    do_plot_pca = False
    if do_plot_pca:
        pca.test_pca(mat=M_unfilled)

    do_plot_fill_rates = False
    if do_plot_fill_rates:
        plot_matrix_fill_rates_by_sector(mat=M_unfilled, df_sectors_=df_sectors)

    do_sk_nmf = False
    if do_sk_nmf:
        try:
            sk_model = NMF(n_components=6, init="random", random_state=0)
            sk_W = sk_model.fit_transform(M_unfilled)
            sk_H = sk_model.components_
        except ValueError as e:
            print("SkLearn NMF does not take missing values !")

    do_imputation_test = False
    if do_imputation_test:
        nmf6_missing, _ = tst.compare_full_sparse_nmf_model(M_unfilled, ncomp=6, sparsity=0.0, nfills=0)
        error_missing = nmf6_missing.get_precision()
        h_missing = nmf6_missing.get_h()
        nmf6_missing_5, _ = tst.compare_full_sparse_nmf_model(M_unfilled, ncomp=6, sparsity=0.0, nfills=5)
        error_missing_5 = nmf6_missing_5.get_precision()
        nmf6_imputed_0, _ = tst.compare_full_sparse_nmf_model(M_filled, ncomp=6, sparsity=0.0, nfills=0)
        error_imputed_0 = nmf6_imputed_0.get_precision()
        h_imputed0 = nmf6_imputed_0.get_h()

    do_features = False
    if do_features:
        sp_nmf6, _ = tst.compare_full_sparse_nmf_model(M_unfilled, ncomp=6, drop_unrelevant=True, do_plot=True)
        ntf6, dense_ntf6 = tst.compare_full_sparse_nmf_model(M_filled, ncomp=6, nblocks=2, drop_unrelevant=True,
                                                             do_plot=False)
        nmf_feat_compo = sp_nmf6.get_features_by_decreasing_loading(min_contribution=0.2)
        ntf_feat_compo = ntf6.get_features_by_decreasing_loading(min_contribution=0.2)
        dense_nmf_feat_compo = sp_nmf6.get_features_by_decreasing_loading(min_contribution=0.2)
        dense_ntf_feat_compo = ntf6.get_features_by_decreasing_loading(min_contribution=0.2)
        sp_nmf6.get_h().to_csv(temp_data_path / "nmf6_h.csv")
        ntf6.get_h().to_csv(temp_data_path / "ntf6_h.csv")
        nmf_feat_compo.to_csv(temp_data_path / "nmf6_feat_compo.csv")
        ntf_feat_compo.to_csv(temp_data_path / "ntf6_feat_compo.csv")
        dense_nmf_feat_compo.to_csv(temp_data_path / "dense_nmf6_feat_compo.csv")
        dense_ntf_feat_compo.to_csv(temp_data_path / "dense_ntf6_feat_compo.csv")
        plot_correspondence(ntf6, df_sectors, pictures_path / "correspondance.pdf")
    nbins = 3
    ncomp = 6

    do_sparsity = False
    # cette étape pour fixer la vacuité optimale dans les modèles
    if do_sparsity:
        for ncomp in [6]:
            df_sparsity_results_nmf = tst.test_nmf_sparsity(M_unfilled, ncomp_=ncomp)
            df_sparsity_results_nmf.to_csv(temp_data_path / f"nmf{ncomp}_sparsity.csv")
            df_sparsity_results_ntf = tst.test_nmf_sparsity(M_filled, nblocks=2, ncomp_=ncomp)
            df_sparsity_results_ntf.to_csv(temp_data_path / f"ntf{ncomp}_sparsity.csv")

    do_comparison = True
    if do_comparison:
        df_comparison = pd.DataFrame(
            index=["nmf", "sparse_nmf", "restricted_nmf",
                   "snmf", "sparse_snmf", "restricted_snmf"],
            columns=["ncomp", "effective_dim", "sparsity", "error", "Clustering entropy", "Matrix entropy", "Delta"],
        )
        nmf6, _ = tst.compare_full_sparse_nmf_model(M_filled, ncomp=6, sparsity=0.0, drop_unrelevant=False,
                                                    do_plot=True)
        df_comparison.loc["nmf", "ncomp"] = nmf6.ncomp
        df_comparison.loc["nmf", "effective_dim"] = nmf6.get_effective_dimension()
        df_comparison.loc["nmf", "sparsity"] = nmf6.get_target_sparsity()
        df_comparison.loc["nmf", "error"] = nmf6.get_precision()
        ce = df_comparison.loc["nmf", "Clustering entropy"] = nmf6.get_avg_diag_entropies(relative=False).iloc[0, 0]
        me = df_comparison.loc["nmf", "Matrix entropy"] = utl.normalized_entropy(nmf6.origin_mat)
        df_comparison.loc["nmf", "Delta"] = ce / me - 1

        sp_nmf6, restricted_nmf6 = tst.compare_full_sparse_nmf_model(M_filled, ncomp=6, sparsity=0.9,
                                                                     drop_unrelevant=True,
                                                                     do_plot=True)
        df_comparison.loc["sparse_nmf", "ncomp"] = sp_nmf6.ncomp
        df_comparison.loc["sparse_nmf", "effective_dim"] = sp_nmf6.get_effective_dimension()
        df_comparison.loc["sparse_nmf", "sparsity"] = sp_nmf6.get_target_sparsity()
        df_comparison.loc["sparse_nmf", "error"] = sp_nmf6.get_precision()
        ce = df_comparison.loc["sparse_nmf", "Clustering entropy"] = \
            sp_nmf6.get_avg_diag_entropies(relative=False).iloc[0, 0]
        me = df_comparison.loc["sparse_nmf", "Matrix entropy"] = \
            utl.normalized_entropy(sp_nmf6.origin_mat)
        df_comparison.loc["sparse_nmf", "Delta"] = ce / me - 1

        # noinspection PyProtectedMember
        df_comparison.loc["restricted_nmf", "ncomp"] = restricted_nmf6._ncomp
        df_comparison.loc["restricted_nmf", "effective_dim"] = restricted_nmf6.get_effective_dimension()
        df_comparison.loc["restricted_nmf", "sparsity"] = restricted_nmf6.get_target_sparsity()
        df_comparison.loc["restricted_nmf", "error"] = restricted_nmf6.get_precision()
        ce = df_comparison.loc["restricted_nmf", "Clustering entropy"] = \
            restricted_nmf6.get_avg_diag_entropies(relative=False).iloc[0, 0]
        me = df_comparison.loc["restricted_nmf", "Matrix entropy"] = \
            utl.normalized_entropy(restricted_nmf6.origin_mat)
        df_comparison.loc["restricted_nmf", "Delta"] = ce / me - 1

        sp_snmf6, _ = tst.compare_full_sparse_nmf_model(M_filled, ncomp=6, nblocks=2, sparsity=0.0, nfills=5,
                                                        drop_unrelevant=False,
                                                        do_plot=True)
        # noinspection PyProtectedMember
        df_comparison.loc["snmf", "ncomp"] = sp_snmf6._ncomp
        df_comparison.loc["snmf", "effective_dim"] = sp_snmf6.get_effective_dimension()
        df_comparison.loc["snmf", "sparsity"] = sp_snmf6.get_target_sparsity()
        df_comparison.loc["snmf", "error"] = sp_snmf6.get_precision()
        ce = df_comparison.loc["snmf", "Clustering entropy"] = \
            sp_snmf6.get_avg_diag_entropies(relative=False).iloc[0, 0]
        me = df_comparison.loc["snmf", "Matrix entropy"] = utl.normalized_entropy(sp_snmf6.origin_mat)
        df_comparison.loc["snmf", "Delta"] = ce / me - 1

        sp_snmf6, restricted_snmf6 = tst.compare_full_sparse_nmf_model(M_filled, ncomp=6, nblocks=2, sparsity=0.9,
                                                                       nfills=5,
                                                                       drop_unrelevant=True, do_plot=True)
        # noinspection PyProtectedMember
        df_comparison.loc["sparse_snmf", "ncomp"] = sp_snmf6._ncomp
        df_comparison.loc["sparse_snmf", "effective_dim"] = sp_snmf6.get_effective_dimension()
        df_comparison.loc["sparse_snmf", "sparsity"] = sp_snmf6.get_target_sparsity()
        df_comparison.loc["sparse_snmf", "error"] = sp_snmf6.get_precision()
        ce = df_comparison.loc["sparse_snmf", "Clustering entropy"] = \
            sp_snmf6.get_avg_diag_entropies(relative=False).iloc[0, 0]
        me = df_comparison.loc["sparse_snmf", "Matrix entropy"] = \
            utl.normalized_entropy(sp_snmf6.origin_mat)
        df_comparison.loc["sparse_snmf", "Delta"] = ce / me - 1

        # noinspection PyProtectedMember
        df_comparison.loc["restricted_snmf", "ncomp"] = restricted_snmf6._ncomp
        df_comparison.loc["restricted_snmf", "effective_dim"] = \
            restricted_snmf6.get_effective_dimension()
        df_comparison.loc["restricted_snmf", "sparsity"] = \
            restricted_snmf6.get_target_sparsity()
        df_comparison.loc["restricted_snmf", "error"] = \
            restricted_snmf6.get_precision()
        ce = df_comparison.loc["restricted_snmf", "Clustering entropy"] = \
            restricted_snmf6.get_avg_diag_entropies(relative=False).iloc[0, 0]
        me = df_comparison.loc["restricted_snmf", "Matrix entropy"] = \
            utl.normalized_entropy(restricted_snmf6.origin_mat)
        df_comparison.loc["restricted_snmf", "Delta"] = ce / me - 1

        df_comparison.to_csv(temp_data_path / "sparse_vs_restricted_models.csv")

    do_compare_diag_entropies = True
    if do_compare_diag_entropies:
        df_delta, diag_nmf, diag_ntf = compare_nmf_ntf_by_entropy(
            M_filled, keep_restricted=True, drop_unr=True,
            nbmin=nbins, nbmax=nbins, ncmin=ncomp, ncmax=ncomp, sp=0.9
        )
        diag_nmf.to_csv(temp_data_path / f"nmf_diag_entropiesc{ncomp}b{nbins}.csv")
        diag_ntf.to_csv(temp_data_path / f"ntf_diag_entropiesc{ncomp}b{nbins}.csv")
        df_delta.to_csv(temp_data_path / f"avg_delta_entropyc{ncomp}b{nbins}.csv")



import numpy as np
import pandas as pd
from . import utilities as utl
from . import nmfmodel as nmf
from . import ntfmodel as ntf
from . import clusterizer as cls
from typing import Union, Tuple
from datetime import datetime

try:
    from adganinference import AdCritic
except ImportError:
    print("Warning: adganinference package was not imported.")


class Imputer:

    def __init__(self,
                 method: str = "median",
                 fill_values: Tuple[float] = None,
                 params: dict = None):

        assert method in {"median", "mean", "values", "kmeans",
                          "nmf.kmeans", "nmf.proxy", "snmf.kmeans", "snmf.proxy"}
        self._method = method
        self._fill_values = fill_values
        self._params = params
        pass

    def __repr__(self):
        return f"Imp_{self._method}_{self._params}"

    def set_ncomp(self, nc: int):
        if self._method in {"nmf.kmeans", "nmf.proxy", "snmf.kmeans", "snmf.proxy"}:
            assert (nc > 1)
            self._params["ncomp"] = nc

    def apply(self,
              mat: Union[np.ndarray, pd.DataFrame],
              mat_name: str = ""):

        if self._method in {"median", "mean", "values", "kmeans"}:
            ngroups = 1
            if self._method == "kmeans":
                ngroups = self._params.get("ngroups", 2)
            filled_mat = utl.fill_missing(mat, self._method, axis=0, ngroups=ngroups)
        elif self._method in {"nmf.kmeans", "nmf.proxy", "snmf.kmeans", "snmf.proxy"}:
            model = None
            ncomp = self._params.get("ncomp", 2)
            nblocks = self._params.get("nblocks", 1)
            nfill_iters = self._params.get("nfill_iters", 1)
            sparsity = self._params.get("sparsity", 0.0)
            use_hals = self._params.get("use_hals", True)
            nmf_type = self._method.split(".")[0]
            fill_type = self._method.split(".")[1]
            if nblocks == 1:
                assert(nmf_type == "nmf")
                model = nmf.NmfModel(mat=mat, name=mat_name, ncomp=ncomp,
                                     regularization="components", leverage="robust",
                                     use_hals=use_hals, max_iter=200)
                model.set_sparsity(sparsity)
            elif nblocks == 2:
                assert (nmf_type == "snmf")
                mat_dup, mat_med = utl.separate_along_median_by_cols(mat)
                model = ntf.NtfModel(
                    mat=mat_dup,
                    name=f"{mat_name}_{ncomp}_snmf",
                    ncomp=ncomp,
                    baseline=mat_med,
                    nblocks=nblocks,
                    regularization="components",
                    leverage="robust",
                    use_hals=use_hals,
                    max_iter=200)
            if model is not None:
                model.test_model(nfill_iters=nfill_iters)
                if fill_type == "proxy":
                    m_proxy = model.get_m_proxy()
                    filled_mat = utl.fill_missing(mat, method="values", fill_values=m_proxy)
                elif fill_type == "kmeans":
                    filled_mat = utl.fill_missing(model.filled_mat, "kmeans", ngroups=ncomp)
            else:
                filled_mat = None

        return filled_mat


class ImputerTester:

    def __init__(self,
                 mat: Union[np.ndarray, pd.DataFrame],
                 name: str,
                 imputer: Imputer,
                 ref_clst: cls.Clusterizer = None,
                 clst: cls.Clusterizer = None,
                 err_func: str = "MSE",
                 **kwargs):

        """

           ImputerTester.__init__:

           Parameters
           ----------
           mat: Union[np.ndarray, pd.DataFrame]: the data to test
           name: test name
           imputer: an Imputer object representing the choosen imputation method
           ref_clst: a Clusterizer object representing a benchmark classification
           clst: a Clusterizer object used to classify the imputed data and give a score
           err_func: a string coding an error function
           # todo: replace the last three parameters by a DataComparator object

           :returns:
           self
           """

        self._mat = mat
        self._name = name
        # self._grps = grps
        self._imputer = imputer
        assert(err_func in ["MSE", "Misclassifieds", "GAN_critic"])
        self._err_func = err_func
        self._ref_clusterizer = ref_clst
        self._clusterizer = clst
        if ref_clst is not None:
            assert (ref_clst.get_nbgroups() == clst.get_nbgroups())
        self._kwargs = kwargs
        self._critic_path = kwargs.get("critic_path", "")
        self._critic_name = kwargs.get("critic_name", "wgan_critic_model")
        self._critic = AdCritic(model_path=self._critic_path,
                                model_name=self._critic_name) if self._err_func.lower() == "gan_critic" else None
        self._run = 0
        self._df_results = pd.DataFrame(columns=["Dataset", "Nb Features", "Nb Obs", "Nb Groups",
                                                 "Imputation strategy",
                                                 "Clustering",
                                                 "Ref Clustering",
                                                 "Missing proportion",
                                                 "Nb Trials",
                                                 "Error Type",
                                                 "Error Min", "Error First decile",
                                                 "Error Median",
                                                 "Error Last decile", "Error Max",
                                                 "Error Stdev",
                                                 "Avg Loss"])
        pass

    def set_imputer(self, imp: Imputer):
        self._imputer = imp

    def set_mat(self, mat):
        self._mat = mat

    def set_name(self, name):
        self._name = name

    def set_clst(self, clst: cls.Clusterizer):
        self._clusterizer = clst

    def set_ref_clst(self, clst: cls.Clusterizer):
        self._ref_clusterizer = clst

    def get_nbgroups(self):
        return self._ref_clusterizer.get_nbgroups() if self._ref_clusterizer is not None else 0

    def run(self,
            missing_props: Tuple[float],
            ntrials: int):

        # imp_mean = Imputer("mean", params={})

        missing_props = sorted(set(missing_props).union({0.0}))
        n_props = len(missing_props)
        run_0 = self._run
        baseline_misclass = 0
        real_scores = []
        if self._err_func.lower() == "gan_critic":
            normed_mat, _, _ = utl.normalize_by_columns(self._mat)
            real_scores = self._critic.predict(normed_mat.values)

        for prop in missing_props:

            self._df_results.loc[self._run, "Dataset"] = self._name
            self._df_results.loc[self._run, "Nb Features"] = len(self._mat.columns)
            self._df_results.loc[self._run, "Nb Obs"] = len(self._mat.index)
            self._df_results.loc[self._run, "Nb Groups"] = self.get_nbgroups()
            self._df_results.loc[self._run, "Imputation strategy"] = self._imputer.__repr__()
            self._df_results.loc[self._run, "Clustering"] = self._clusterizer.__repr__()
            self._df_results.loc[self._run, "Ref Clustering"] = self._ref_clusterizer.__repr__()
            self._df_results.loc[self._run, "Error Type"] = self._err_func
            self._df_results.loc[self._run, "Missing proportion"] = prop
            self._df_results.loc[self._run, "Nb Trials"] = ntrials if prop > 0 else 1
            df_trials = pd.DataFrame(index=range(ntrials),
                                     columns=["Error"])
            trials = range(ntrials) if prop > 0.0 else range(1)
            rndg = np.random.default_rng(seed=42)
            # Générer les matrices remplies après censure aléatoire
            for itrial in trials:
                if prop > 0:
                    m_cens = utl.censor_data(self._mat, prop, inplace=False, randomgen=rndg)
                else:
                    m_cens = self._mat.copy()
                filled_m = self._imputer.apply(m_cens, self._name).astype(float)
                holes = m_cens.apply(lambda x: pd.isnull(x)).astype(int)
                if self._err_func.lower() != "misclassifieds":
                    if "cluster" in filled_m.columns:
                        filled_m = filled_m.drop("cluster", axis=1)
                if self._err_func.lower() == "mse":
                    df_trials.loc[itrial, "ErrorOnMissings"] = utl.relative_squared_error(np.multiply(self._mat, holes),
                                                                                          np.multiply(filled_m, holes),
                                                                                          self._mat)
                    df_trials.loc[itrial, "Error"] = utl.relative_squared_error(self._mat, filled_m)
                elif self._err_func.lower() == "misclassifieds":
                    _, df_trials.loc[itrial, "Error"] = utl.get_nb_misclassifieds(filled_m,
                                                                                  self._clusterizer,
                                                                                  self._ref_clusterizer) \
                                                        if self._clusterizer is not None \
                                                        else np.nan
                elif self._err_func.lower() == "gan_critic":
                    normed_filled_mat, _, _ = utl.normalize_by_columns(filled_m)
                    scores = self._critic.predict(normed_filled_mat)
                    # todo: compute a loss between the scores obtained on real data and the scores on imputed data
                    df_trials.loc[itrial, "Error"] = np.mean(real_scores - scores)

            errors = df_trials["Error"].astype(float)
            missings_errs = df_trials["ErrorOnMissings"].astype(float)
            self._df_results.loc[self._run, "Error Min"] = np.nanmin(errors)
            self._df_results.loc[self._run, "Error First decile"] = np.nanpercentile(errors, q=10)
            self._df_results.loc[self._run, "Error Median"] = np.nanmedian(errors)
            self._df_results.loc[self._run, "ErrorMissings Median"] = np.nanmedian(missings_errs)
            self._df_results.loc[self._run, "Error Last decile"] = np.nanpercentile(errors, q=90)
            self._df_results.loc[self._run, "Error Max"] = np.nanmax(errors)
            self._df_results.loc[self._run, "Error Stdev"] = np.nanstd(errors)
            self._df_results.loc[self._run, "Error Avg"] = np.nanmean(errors)
            baseline_misclass = self._df_results.loc[run_0, "Error Median"]
            self._run += 1

            print(f"Done {self._imputer}_({prop}) on {self._name}")

        for iprop in range(n_props):
            self._df_results.loc[run_0 + iprop, "Avg Loss"] = self._df_results.loc[run_0 + iprop, "Error Avg"] - \
                baseline_misclass

        return

    def output_results(self, outpath):

        self._df_results.to_csv(outpath / f"Imputer _{self._name}_{datetime.today().strftime('%Y-%m-%d-%H:%M')}.csv")

    pass









# import itertools
# import os

import numpy as np
import pandas as pd
from adlinear import nmfmodel as nmf
from randomgenerators import randomgenerators as rng
from transparentpath import TransparentPath
import dotenv
dotenv.load_dotenv()

if __name__ == "__main__":

    train_f_name = f"wfold_scree_plots"
    train_file = rng.RandomNamedFile(dirpath=TransparentPath(),
                                     root_name=train_f_name,
                                     suffix="csv")

    rand_norms = True
    df_mini_scree_plots = pd.DataFrame(index=[], columns=[])
    for itrial in range(100):
        nb_clusters = np.random.randint(low=5, high=40)
        min_corr = np.random.uniform(low=0.75, high=0.95)
        max_corr = np.random.uniform(low=0.05, high=0.25)
        h_avg_csize = np.random.randint(low=10, high=20)
        w_avg_csize = np.random.randint(low=10, high=100)
        eps = np.random.uniform(low=0.0, high=0.25)

        generated_M = rng.generate_nmf_reconstruction(n_comp=nb_clusters,
                                                      min_intra_corr=min_corr,
                                                      max_inter_corr=max_corr,
                                                      h_avg_clust_size=h_avg_csize,
                                                      w_avg_clust_size=w_avg_csize,
                                                      epsilon=eps,
                                                      random_norms=rand_norms)
        rnstr = "RandNorms" if rand_norms else "ConstNorms"
        # generated_M.to_csv(res_path / "random_nmf" / f"M_t{itrial}_nc{nb_clusters}_corrmin{round(min_corr,2)}_"
        #                                f"corrmax{round(max_corr,2)}_noise{round(eps,2)}_{rnstr}.csv",
        #                    float_format="%.4f")
        df_scree_plot = nmf.generate_scree_plot(generated_M, ncmin=3, ncmax=45)
        df_mini_scree_plots = nmf.add_miniscree_plots_from_scree_plot(df_mini_scree_plots, df_scree_plot, nb_clusters)

    train_file.filepath.write(df_mini_scree_plots)




import pandas as pd
import os
import numpy as np
# import math

# from sklearn.decomposition import NMF

# from adlinear import pca
from adlinear import utilities as utl
from adlinear import nmfmodel as nmf
from adlinear import testers as tst
from adlinear import features_selector as fs
from randomgenerators import randomgenerators as rng
# from adlinear import ntfmodel as ntf
from sklearn.svm import SVC, SVR
from sklearn.linear_model import ElasticNet

# import matplotlib
# import matplotlib.pyplot as plt
# import matplotlib.ticker as mtick
# import seaborn as sns
from pathlib import Path




import numpy as np
import pandas as pd
from . import utilities as utl
from . import nmfmodel as nmf
from . import ntfmodel as ntf
from datetime import datetime

from pathlib import Path
from sklearn import metrics

class FeatureSelector:
    def __init__(self,
                 name="FS",
                 method: str = "random",
                 **kwargs):
        method = method.lower()
        assert(method in ["random", "nmf", "snmf",
                          "elastic_net",
                          "saved_features",
                          "score_based",
                          "none"])
        self._name = name
        self._method = method
        self._kwargs = kwargs
        self._mat = None
        self._submat = None
        self._dropped_features = None

    def get_name(self):
        return self._name

    def get_fuzzy_rank(self):
        return self._kwargs.get("fuzzy_rank", 0)

    def get_what_to_drop(self):
        what_to_drop = self._kwargs.get("what_to_drop", "null")
        assert what_to_drop in ["null", "fuzzy", "both"]
        return what_to_drop

    def get_prop_select(self):
        return self._kwargs.get("prop_select", 0.5)

    def get_nfeat_select(self):
        return self._kwargs.get("nfeat_select", 0)

    def get_ncomp(self):
        return self._kwargs.get("ncomp", 2)

    def get_nblocks(self):
        return self._kwargs.get("nblocks", 1)

    def get_sparsity(self):
        return self._kwargs.get("sparsity", 1.0)

    def get_alpha(self):
        return self._kwargs.get("alpha", 1.0)

    def get_l1_ratio(self):
        return self._kwargs.get("l1_ratio", 0.5)

    def get_submat(self):
        return self._submat

    def get_dropped_features(self):
        return self._dropped_features

    def get_base_nb_features(self):
        return len(self._mat.columns)

    def get_reduced_nb_features(self):
        return len(self._submat.columns)

    def apply(self, mat, **kwargs):
        self._mat = pd.DataFrame(mat)
        name = self._kwargs.get("name", "dataset")

        if self._method == "random":
            nfeat_select = self.get_nfeat_select()
            nfeat = len(self._mat.columns)
            if nfeat_select == 0:
                prop = self.get_prop_select()
                nfeat_select = int(prop * nfeat)
            rng = np.random.default_rng()
            kept_cols = rng.choice(range(nfeat), nfeat_select, replace=False)
            self._submat = self._mat.copy().iloc[:, kept_cols]
            dropped_cols = set(self._mat.columns).difference(self._mat.columns[kept_cols])
            self._dropped_features = dropped_cols

        elif self._method in ["nmf", "snmf"]:
            nbl = self.get_nblocks()
            is_separative = nbl > 1
            what = self.get_what_to_drop()
            rank = self.get_fuzzy_rank()
            ncomp = self.get_ncomp()
            drop_null = what in ["null"]
            drop_fuzzy = what in ["fuzzy"]

            if not is_separative:
                nmf_model = nmf.NmfModel(mat=self._mat, ncomp=self.get_ncomp(),
                                         name=name, regularization="components")
            else:
                nmf_model = ntf.NtfModel(mat=self._mat, ncomp=self.get_ncomp(), unfolded=False,
                                         nblocks=2,
                                         name=name, regularization="components")
            nmf_model.set_sparsity(self.get_sparsity())
            to_drop_features = []
            if drop_null:
                to_drop_features = nmf_model.get_null_features()
            elif drop_fuzzy:
                to_drop_features = nmf_model.get_fuzzy_features(rank)

            self._dropped_features = to_drop_features
            self._submat = self._mat.copy().drop(to_drop_features, axis=1) \
                if to_drop_features and len(to_drop_features) > 0 \
                else self._mat.copy()


        elif self._method == "none":
            self._submat = self._mat.copy()
            self._dropped_features = None

        return self._submat, self._dropped_features

    def __repr__(self):
        return f"{self._name}:{self._method}_{self._kwargs.__repr__()}"

    def output_features(self, outpath):
        df_features = pd.DataFrame(data=self._submat.columns)
        df_features.to_csv(outpath / f"_{self.__repr__()}.csv")


class FeatureSelectorTester:

    def __init__(self, name="FST", predictive_template=None, fselectors=[], dataset_name="Data", training_set=None,
                 validation_set=None, test_set=None, training_labels=None, validation_labels=None, test_labels=None,
                 classifying_y=False, **kwargs):

        def safe_copy(dataset):
            return dataset.copy() if dataset is not None else None

        self._name = name
        self._predictive_template = predictive_template
        self._full_model = None
        self._reduced_model = None
        self._fselectors = fselectors
        self._dataset_name = dataset_name
        self._full_training_set = safe_copy(training_set)
        self._reduced_training_set = safe_copy(training_set)
        self._training_labels = safe_copy(training_labels)

        self._full_validation_set = safe_copy(validation_set)
        self._reduced_validation_set = safe_copy(validation_set)
        self._validation_labels = safe_copy(validation_labels)

        self._full_test_set = safe_copy(test_set)
        self._reduced_test_set = safe_copy(test_set)
        self._test_labels = safe_copy(test_labels)
        self._has_binary_y = classifying_y

        self._kwargs = kwargs
        self._run = 0
        self._df_results = pd.DataFrame(columns=["Dataset", "Model", "Feature Selection", "Nb Features",
                                                 "Base Train Score",
                                                 "Base Test Score", "Reduced Nb Features", "Reduced Train Score",
                                                 "Reduced Test Score"])
        pass

    def set_predictive_model(self, predictive_template, **kwargs):
        self._predictive_template = predictive_template
        self._kwargs = kwargs

    def get_feature_selectors(self):
        return self._fselectors

    def set_feature_selectors(self, fselectors):
        self._fselectors = fselectors

    def get_feature_selector(self, fselector, i=0):
        return self._fselectors[i]

    def set_feature_selector(self, fselector, i=0):
        self._fselectors[i] = fselector

    def get_nb_selectors(self):
        return len(self._fselectors)

    def run(self, random_tests=0):
        # instantiation du modèle
        mdl = self._predictive_template(**self._kwargs) if len(self._kwargs) > 0 else self._predictive_template()
        if mdl is None:
            return
        self._full_model = mdl
        # apprentissage sur les données d'apprentissage complètes
        mdl.fit(self._full_training_set.values, self._training_labels.values)
        y_full_train = mdl.predict(self._full_training_set.values)
        if self._has_binary_y:
            y_full_train = np.round(y_full_train, 0)
        # calcul des scores d'apprentissage
        score_function = metrics.accuracy_score if self._has_binary_y else metrics.mean_absolute_error
        score_train_full = score_function(y_true=self._training_labels.values,
                                          y_pred=y_full_train)
        # score de validation pour lequel les y sont connus
        y_full_val = mdl.predict(self._full_validation_set.values)
        if self._has_binary_y:
            y_full_val = np.round(y_full_val, 0)
        score_val_full = score_function(y_true=self._training_labels.values,
                                        y_pred=y_full_val)
        self._df_results.loc[self._run, "Dataset"] = self._dataset_name
        self._df_results.loc[self._run, "Model"] = mdl.__repr__()
        self._df_results.loc[self._run, "Feature Selection"] = self._fselectors[0].__repr__() \
            if self.get_nb_selectors() == 1 else [fsel.get_name() for fsel in self._fselectors].__repr__()
        self._df_results.loc[self._run, "Nb Features"] = len(self._full_training_set.columns)
        self._df_results.loc[self._run, "Base Train Score"] = score_train_full
        self._df_results.loc[self._run, "Base Validation Score"] = score_val_full

        # création d'un modèle réduit
        reduced_mdl = self._predictive_template(**self._kwargs) if len(self._kwargs) > 0 \
            else self._predictive_template()
        self._reduced_model = reduced_mdl
        # parcours des stratégies d'élimination de features

        self._reduced_training_set = self._full_training_set.copy()
        self._reduced_validation_set = self._full_validation_set.copy()
        self._reduced_test_set = self._full_test_set.copy()
        not_empty = True
        for selector in self._fselectors:
            # calcul des features à éliminer
            reduced_training_set, dropped_features = selector.apply(self._reduced_training_set)
            not_empty = len(reduced_training_set.columns) > 0
            if not_empty:
                self._reduced_training_set = utl.drop_features(self._reduced_training_set, dropped_features)

                self._reduced_validation_set = utl.drop_features(self._reduced_validation_set, dropped_features)

                self._reduced_test_set = utl.drop_features(self._reduced_test_set, dropped_features)
            else:
                break

        if not_empty:
            # prédiction sur les ensembles réduits
            reduced_mdl.fit(self._reduced_training_set.values, self._training_labels.values)
        # calcul des scores sur les ensembles d'apprentissage et de validation
        # le test n'est pas en général disponible
        y_reduced_train = reduced_mdl.predict(self._reduced_training_set.values)
        if self._has_binary_y:
            y_reduced_train = np.round(y_reduced_train, 0)
        score_tr_red = score_function(y_true=self._training_labels,
                                      y_pred=y_reduced_train) if not_empty else 0
        y_reduced_val = reduced_mdl.predict(self._reduced_validation_set.values)
        if self._has_binary_y:
            y_reduced_val = np.round(y_reduced_val, 0)
        score_val_red = score_function(y_true=self._training_labels,
                                       y_pred=y_reduced_val) if not_empty else 0

        self._df_results.loc[self._run, "Reduced Nb Features"] = len(self._reduced_training_set.columns)
        self._df_results.loc[self._run, "Reduced Train Score"] = score_tr_red
        self._df_results.loc[self._run, "Reduced Validation Score"] = score_val_red

        self._run += 1

    def run_random_selections(self, nb_rand_trials=0):
        red_nbf = self.get_reduced_nb_features()
        for _ in range(nb_rand_trials):
            rand_fsel = FeatureSelector(method="random", nfeat_select=red_nbf)
            self.set_feature_selectors([rand_fsel])
            self.run()

    def get_reduced_nb_features(self):
        return 0 if self._reduced_training_set is None else len(self._reduced_training_set.columns)

    def output_all_results(self, outpath):
        self._df_results.to_csv(outpath / f"{self._name}_{datetime.today().strftime('%Y-%m-%d')}.csv")

    def output_current_model(self, dataset_name, outpath):
        if self._full_model is None:
            return
        y_pred_full = self._full_model.predict(self._full_test_set.values)
        np.savetxt(Path(outpath) / f"{dataset_name}_full_test.predict", y_pred_full, fmt="%.0f")
        np.savetxt(Path(outpath) / f"{dataset_name}_full.feat", self._full_test_set.columns, fmt="%s")
        y_pred_reduced = self._reduced_model.predict(self._reduced_test_set.values)
        np.savetxt(Path(outpath) / f"{dataset_name}_{self._fselectors}_reduced_test.predict",
                   y_pred_reduced, fmt="%.0f")
        np.savetxt(Path(outpath) / f"{dataset_name}_{self._fselectors}_reduced.feat",
                   self._reduced_test_set.columns, fmt="%s")


if __name__ == "__main__":

    fs = FeatureSelector("random", prop=2)
